"""
Unit tests to validate the performance of the BrainVision reader

This class tests measures the performance of reading (chunked or non-chunked, standard IO and Memmap):
   - a single channel (all samples)
   - the full set (all channels, all samples)
   - 100 000 samples (all channels)
   - multiple (two) ranges of 1000 samples (all channels)

The performance results are stored as a pickle (.pdat) and a Matlab (.mat) file and optionally can be visualized
using the 'matlab/visualize_perfTestResults.m' script


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest
import pickle
import numpy as np
from datetime import datetime
from ieegprep.utils.console import ConsoleColors
from ieegprep.utils.misc import clear_virtual_cache
from ieegprep.fileio.BrainVisionReader import BrainVisionReader

class TestFileIO(unittest.TestCase):
    """
    Validate the performance of the BrainVision reader
    """

    data_orientation = 'multiplexed'
    #data_orientation = 'vectorized'

    bv_data_path = 'D:\\BIDS_erdetect\\sub-BV\\ses-1\\ieeg\\sub-BV_ses-1_ieeg.vhdr'
    #bv_data_path = os.path.expanduser('~/Documents/ERDetect_perf/sub-BV/ses-1/ieeg/sub-BV_ses-1_ieeg.vhdr')


    def test_fileio_bv_perf(self):
        """
        Test reader performance
        """
        from ieegprep.utils.misc import time_func

        uncached_read_repetitions = 20
        cached_read_repetitions = 100

        # define tests
        tests = dict()
        tests['single_channel'] = dict()
        tests['single_channel']['uncached_read_repetitions'] = uncached_read_repetitions
        tests['single_channel']['cached_read_repetitions'] = cached_read_repetitions
        tests['single_channel']['channels'] = ('CH07',)
        tests['single_channel']['start_sample'] = 0
        tests['single_channel']['end_sample'] = -1
        tests['single_channel']['conditions'] = []
        tests['single_channel']['conditions'].append((True, True))              # memmap & chunked
        tests['single_channel']['conditions'].append((True, False))             # memmap & non-chunked
        tests['single_channel']['conditions'].append((False, True))             # stdIO & chunked
        if self.data_orientation.lower() == 'vectorized':                           # (skip on multiplexed)
            tests['single_channel']['conditions'].append((False, False))        # stdIO & non-chunked

        tests['full'] = dict()
        tests['full']['uncached_read_repetitions'] = uncached_read_repetitions
        tests['full']['cached_read_repetitions'] = cached_read_repetitions
        tests['full']['channels'] = ()
        tests['full']['start_sample'] = 0
        tests['full']['end_sample'] = -1
        tests['full']['conditions'] = []
        tests['full']['conditions'].append((True, True))                # memmap & chunked
        tests['full']['conditions'].append((True, False))               # memmap & non-chunked
        tests['full']['conditions'].append((False, True))               # stdIO & chunked
        tests['full']['conditions'].append((False, False))              # stdIO & non-chunked

        tests['range_100k'] = dict()
        tests['range_100k']['uncached_read_repetitions'] = uncached_read_repetitions
        tests['range_100k']['cached_read_repetitions'] = cached_read_repetitions
        tests['range_100k']['channels'] = ()
        tests['range_100k']['start_sample'] = 1000
        tests['range_100k']['end_sample'] = 101000
        tests['range_100k']['conditions'] = []
        tests['range_100k']['conditions'].append((True, True))               # memmap & chunked
        tests['range_100k']['conditions'].append((True, False))              # memmap & non-chunked
        tests['range_100k']['conditions'].append((False, True))              # stdIO & chunked
        tests['range_100k']['conditions'].append((False, False))             # stdIO & non-chunked

        tests['multi_range'] = dict()
        tests['multi_range']['uncached_read_repetitions'] = uncached_read_repetitions
        tests['multi_range']['cached_read_repetitions'] = cached_read_repetitions
        tests['multi_range']['channels'] = ()
        tests['multi_range']['start_sample'] = [1000, 101000]
        tests['multi_range']['end_sample'] = [2000, 102000]
        tests['multi_range']['conditions'] = []
        tests['multi_range']['conditions'].append((True, True))          # memmap & chunked
        tests['multi_range']['conditions'].append((True, False))         # memmap & non-chunked
        tests['multi_range']['conditions'].append((False, True))         # stdIO & chunked
        tests['multi_range']['conditions'].append((False, False))        # stdIO & non-chunked

        # check if orientation matches
        hdr = BrainVisionReader.bv_read_header(self.bv_data_path)
        self.assertEqual(hdr['data_orientation'], self.data_orientation.upper(), 'Data orientation mismatch. Make sure the \'data_orientation\' variable in the test (' + self.data_orientation.upper() + ') matches the validation set\'s \'DataOrientation\' field (' + hdr['data_orientation'] + ')')

        # retrieve platform data
        import platform
        platform_info = dict()
        platform_info['platform'] = platform.system()
        platform_info['platform-release'] = platform.release()
        platform_info['platform-version'] = platform.version()
        platform_info['architecture'] = platform.machine()
        platform_info['processor'] = platform.processor()


        #
        # perform tests
        #

        # loop over the tests
        for test_name, test in tests.items():
            ConsoleColors.print_green('Test: ' + test_name)

            tests[test_name]['cond_results_uncached'] = []
            tests[test_name]['cond_results_cached'] = []

            # loop over the conditions
            for condition in test['conditions']:
                text_reading_type = 'memmap' if condition[0] else 'stdIO'
                text_reading_chunked = 'chunked' if condition[1] else 'non-chunked'
                print('    - Condition - ' + text_reading_type + ' - ' + text_reading_chunked)

                # test with cache clearing
                if test['uncached_read_repetitions'] > 0:
                    print('      - Uncached (with cache-clearing)')

                    # test un-cached
                    # Note: adding 'clear_virtual_cache' as pre_func will clear the virtual cache (page memory)
                    #       before reading (but will not be taken into account in the performance measurement)
                    mean, std, range, times = time_func(self._test_reader_read, clear_virtual_cache, test['uncached_read_repetitions'], self.bv_data_path,
                                                        use_memmap=condition[0], chunked_read=condition[1],
                                                        channels=test['channels'], start_sample=test['start_sample'], end_sample=test['end_sample'])

                    print('          Mean: ' + str(round(mean, 2)) + ' - std: ' + str(round(std, 2)) + ' - range: ' + str(round(range[0], 2)) + '-' + str(round(range[1], 2)))
                    tests[test_name]['cond_results_uncached'].append(np.asarray([mean, std, range, times], dtype=object))

                # test while cached
                if test['cached_read_repetitions'] > 0:
                    print('      - Cached')

                    # once to make sure the data is cached
                    self._test_reader_read(self.bv_data_path, use_memmap=condition[0], chunked_read=condition[1],
                                           channels=test['channels'], start_sample=test['start_sample'], end_sample=test['end_sample'])

                    # test cached
                    mean, std, range, times = time_func(self._test_reader_read, None, test['cached_read_repetitions'], self.bv_data_path,
                                                        use_memmap=condition[0], chunked_read=condition[1],
                                                        channels=test['channels'], start_sample=test['start_sample'], end_sample=test['end_sample'])

                    print('          Mean: ' + str(round(mean, 2)) + ' - std: ' + str(round(std, 2)) + ' - range: ' + str(round(range[0], 2)) + '-' + str(round(range[1], 2)))
                    tests[test_name]['cond_results_cached'].append(np.asarray([mean, std, range, times], dtype=object))

                # clear the virtual cache at the end of the read condition
                clear_virtual_cache()


        #
        # Store the results to disk
        #

        # build the filename
        os_text = 'unk'
        if platform_info['platform'].lower() in ('win32', 'windows'):
            os_text = 'win'
        elif platform_info['platform'].lower() in ('linux', 'linux2'):
            os_text = 'lnx'
        elif platform_info['platform'].lower() == 'darwin':
            os_text = 'mac'
        output_file = 'results_bv_' + os_text + '_' + self.data_orientation + '__' + datetime.now().strftime("%Y%m%d_%H%M%S")

        # as pickle
        with open('./' + output_file + '.pdat', 'wb') as output_handle:
            pickle.dump([platform_info, self.data_orientation, tests], output_handle)

        # as matlab
        import scipy
        output = dict()
        output['platform'] = platform_info
        output['data_orientation'] = self.data_orientation
        output['tests'] = tests
        scipy.io.savemat('./' + output_file + '.mat', output)

        #
        ConsoleColors.print_green('Tests complete')


    @staticmethod
    def _test_reader_read(data_path, use_memmap, chunked_read, channels, start_sample=0, end_sample=-1):
        """
        Perform a single read with the reader

        Args:
            data_path (str):
            use_memmap (bool):              Whether to use numpy's memmap (which wraps around mmap) while reading the data.
            chunked_read (bool):            Whether to read or transfer the data in chunks
            channels (str/list/tuple):      The names of the channels to return the signal data from.
            start_sample (int or list):     The start-point in time (in samples) to start reading from (0-based)
            end_sample (int or list):       The sample to end the reading (0-based)

        """

        hdr, data_io = BrainVisionReader.bv_read_data(data_path,
                                                      use_memmap=use_memmap, chunked_read=chunked_read,
                                                      channels=channels, start_sample=start_sample, end_sample=end_sample,
                                                      unit='uV')
        del hdr, data_io


if __name__ == '__main__':
    unittest.main()



"""
Unit tests to validate the BrainVision reader

This class tests reading (chunked or non-chunked, standard IO and Memmap):
   - a single channel (all samples)                       --> test01_fileio_bv_validate_data_single_channel
   - the full set (all channels, all samples)             --> test02_fileio_bv_validate_data_fullset
   - 100 000 samples (all channels)                       --> test03_fileio_bv_validate_data_100k_samples
   - multiple (two) ranges of 1000 samples (all channels) --> test04_fileio_bv_validate_data_multi_range

Each of these tests uses the BrainVision reader to read data and validates the read data against:
    - a specific Matlab data file (.mat) that has been produced by the 'matlab/generate_bv_verification_sets.m' script,
      which in turn uses Fieldtrip to read the data
    - the same subset of data from the full dataset read using MNE


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest
from ieegprep.utils.console import ConsoleColors


class TestFileIOBVData(unittest.TestCase):
    """
    Validate the BrainVision reader output values against the Matlab (fieldtrip) and MNE reader output values
    """

    #
    # Configuration
    #

    data_orientation = 'multiplexed'
    #data_orientation = 'vectorized'

    matlab_data_path = 'D:\\BIDS_erdetect\\bv_'
    bv_data_path = 'D:\\BIDS_erdetect\\sub-BV\\ses-1\\ieeg\\sub-BV_ses-1_ieeg.vhdr'
    #matlab_data_path = os.path.expanduser('~/Documents/ERDetect_perf/bv_')
    #bv_data_path = os.path.expanduser('~/Documents/ERDetect_perf/sub-BV/ses-1/ieeg/sub-BV_ses-1_ieeg.vhdr')

    # set part of the filename based on the configuration
    orient_path = 'mplex_' if data_orientation.lower() == 'multiplexed' else 'vectd_'


    def test01_fileio_bv_validate_data_single_channel(self):
        # validate single channel

        #
        # matlab data
        #
        import scipy.io
        mat = scipy.io.loadmat(self.matlab_data_path + self.orient_path + 'allsamples_CH07.mat')

        # Note: When reading a single channel from multiplexed BV data, stdIO non-chunked reading will automatically
        #       switch to chunked reading to prevent slow per sample reading. As such, stdIO non-chunked reading
        #       is skipped here because it wouldn't really get tested
        skip_stdIO_non_chunked = self.data_orientation.lower() == 'multiplexed'
        self._test_reader_output_matrix(mat['dat'], 0, 'Matlab single channel', channels=('CH07',), skip_stdIO_non_chunked=skip_stdIO_non_chunked)
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_brainvision
        mne_raw = read_raw_brainvision(self.bv_data_path, eog=[], misc=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here
        try:
            channel_idx = mne_raw.info['ch_names'].index('CH07')
        except ValueError:
            self.assertEqual(0, 1, 'Could not find channel \'' + 'CH07' + '\' in the dataset')

        self._test_reader_output_matrix(np.array([mne_raw._data[channel_idx, :]]), np.finfo(np.float32).eps, 'MNE single channel', channels=('CH07',), skip_stdIO_non_chunked=skip_stdIO_non_chunked)
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Single channel test successful')


    def test02_fileio_bv_validate_data_fullset(self):
        # full dataset - all samples - all channels

        #
        # matlab data
        #
        import mat73
        mat = mat73.loadmat(self.matlab_data_path + self.orient_path + 'allsamples_allchannels.mat')
        self._test_reader_output_matrix(mat['dat'], 0, 'Matlab full dataset')
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_brainvision
        mne_raw = read_raw_brainvision(self.bv_data_path, eog=[], misc=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        self._test_reader_output_matrix(mne_raw._data, np.finfo(np.float32).eps, 'MNE full dataset')
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Full dataset test successful')
        

    def test03_fileio_bv_validate_data_100k_samples(self):
        # 100 000 samples - all channels

        #
        # matlab data
        #
        import scipy.io
        mat = scipy.io.loadmat(self.matlab_data_path + self.orient_path + '100ksamples.mat')
        self._test_reader_output_matrix(mat['dat'], 0, 'Matlab 100k samples', start_sample=1000, end_sample=101000)
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_brainvision
        mne_raw = read_raw_brainvision(self.bv_data_path, eog=[], misc=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        self._test_reader_output_matrix(mne_raw._data[:, 1000:101000], np.finfo(np.float32).eps, 'MNE 100k samples', start_sample=1000, end_sample=101000)
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('100k samples test successful')
    

    """
    def test03b_fileio_bv_validate_data_100k_samples(self):
        # sliding ranges - single channel

        skip_stdIO_non_chunked = self.data_orientation.lower() == 'multiplexed'
        from ieegprep.fileio.BrainVisionReader import BrainVisionReader
        hdr = BrainVisionReader.bv_read_header(self.bv_data_path)

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_brainvision
        mne_raw = read_raw_brainvision(self.bv_data_path, eog=[], misc=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        
        # test slide sample over a small number of blocks (1 and 2)
        for start_sample in range(0, 1000, 1):
            for end_sample in range(1000, 400, -1):
                if start_sample < end_sample:
                    print('Test - start sample: ' + str(start_sample) + ' - end sample:' + str(end_sample))
                    self._test_reader_output_matrix(np.array([mne_raw._data[1, start_sample:end_sample]]), np.finfo(np.float32).eps,
                                                    'MNE sliding ranges',
                                                    start_sample=start_sample, end_sample=end_sample, channels=('CH02',),
                                                    skip_stdIO_non_chunked=skip_stdIO_non_chunked, silent=True)

        # test slide over all sample possibilities where the start is between 0 and 10, and the end is between end-10 and end
        for start_sample in range(0, 10, 1):
            for end_sample in range(hdr['number_of_samples'], hdr['number_of_samples'] - 10, -1):
                print('Test - start sample: ' + str(start_sample) + ' - end sample:' + str(end_sample))
                self._test_reader_output_matrix(np.array([mne_raw._data[1, start_sample:end_sample]]), np.finfo(np.float32).eps,
                                                'MNE sliding ranges',
                                                start_sample=start_sample, end_sample=end_sample, channels=('CH02',),
                                                skip_stdIO_non_chunked=skip_stdIO_non_chunked, silent=True)

        # test slide in step of 155 over the sample possibilities where the start is between 0 and 1000, and the end is between end-1000 and end
        for start_sample in range(0, 1000, 155):
            for end_sample in range(hdr['number_of_samples'], hdr['number_of_samples'] - 1000, -155):
                print('Test - start sample: ' + str(start_sample) + ' - end sample:' + str(end_sample))
                self._test_reader_output_matrix(np.array([mne_raw._data[1, start_sample:end_sample]]), np.finfo(np.float32).eps,
                                                'MNE sliding ranges',
                                                start_sample=start_sample, end_sample=end_sample, channels=('CH02',),
                                                skip_stdIO_non_chunked=skip_stdIO_non_chunked, silent=True)


        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Sliding ranges test successful')
    """

    def test04_fileio_bv_validate_data_multi_range(self):
        # multiple ranges - all channels

        #
        # matlab data
        #
        import scipy.io
        mat = scipy.io.loadmat(self.matlab_data_path + self.orient_path + 'multiranges.mat')
        self._test_reader_output_matrix(mat['dat'], 0, 'Matlab multiple sample ranges', start_sample=(1000, 101000), end_sample=(2000, 102000))
        del mat
        

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_brainvision
        mne_raw = read_raw_brainvision(self.bv_data_path, eog=[], misc=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        self._test_reader_output_matrix(np.stack((mne_raw._data[:, 1000:2000], mne_raw._data[:, 101000:102000]), axis=-1), np.finfo(np.float32).eps, 'MNE multiple sample ranges', start_sample=(1000, 101000), end_sample=(2000, 102000))
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Multiple sample ranges test successful')


    def _test_reader_output_matrix(self, mat, diff_tolerance, title, channels=None, start_sample=0, end_sample=-1, skip_stdIO_non_chunked=False, silent=False):
        """
        Test the output values of four different types of reading (chunked/non-chucked and stdIO/memmap) against
        a given array of values

        mat (nparray):                      Array of values to test against
        diff_tolerance (float):             The difference between the output values that is tolerated.
                                            The precision of the float values and manipulations from MNE cause slight
                                            differences (~.00000000001 for MNE) in the outputs, this tolerance value
                                            allows for less precise testing
        title (str):                        The title of the test
        skip_stdIO_non_chunked (bool):      Whether to skip the standard IO non-chunked reading test. In a multiplexed
                                            dataset, when either not reading all channels or not reading them in
                                            the same order as in which they are stored, then each samples is read separately
                                            in nested loops. Reading per sample is very slow, and this type of reading
                                            is automatically switched to chunked reading. Therefore non-chunked reading
                                            under these conditions might not actually happen and can be skipped

        """
        from ieegprep.fileio.BrainVisionReader import BrainVisionReader
        #import numpy as np

        #
        def assert_results(hdr, data_io, type):
            self.assertEqual(hdr['data_orientation'], self.data_orientation.upper(), 'Data orientation mismatch. Make sure the \'data_orientation\' variable in the test (' + self.data_orientation.upper() + ') matches the validation set\'s \'DataOrientation\' field (' + hdr['data_orientation'] + ')')

            # find the differences in data and report on the largest difference
            diff = abs(mat - data_io)
            #max_diff = diff.max()
            #if max_diff > 0:
            #    max_diff_index = np.unravel_index(diff.argmax(), diff.shape)
            #    if not silent:
            #        print(' - A maximum difference of ' + str(max_diff) + ' was found at channel ' + hdr['channel_names'][max_diff_index[0]] + ' (index ' + str(max_diff_index[0]) + '). At least at sample: ' + str(max_diff_index[1]) + ' (read value: ' + str(data_io[max_diff_index[0], max_diff_index[1]]) + '; matlab/MNE value: ' + str(mat[max_diff_index[0], max_diff_index[1]]) + ')')

            # assert
            self.assertEqual((diff <= diff_tolerance).sum(), data_io.size, ('Validation of \'' + self.matlab_data_path + '*.mat\' failed while testing ' + title + ' (' + hdr['data_orientation'].lower() + ') and reading with ' + type))

        #
        # stdIO, chunked
        #
        if not silent:
            ConsoleColors.print_green('Testing ' + title + ' with chunked stdIO')
        hdr, data_io = BrainVisionReader.bv_read_data(self.bv_data_path, channels=channels,
                                                      start_sample=start_sample, end_sample=end_sample, unit='uV',
                                                      use_memmap=False, chunked_read=True)
        assert_results(hdr, data_io, 'chunked stdIO')
        del hdr, data_io

        #
        # stdIO, non-chunked
        #
        if not skip_stdIO_non_chunked:
            if not silent:
                ConsoleColors.print_green('Testing ' + title + ' with non-chunked stdIO')
            hdr, data_io = BrainVisionReader.bv_read_data(self.bv_data_path, channels=channels,
                                                          start_sample=start_sample, end_sample=end_sample, unit='uV',
                                                          use_memmap=False, chunked_read=False)
            assert_results(hdr, data_io, 'non-chunked stdIO')
            del data_io

        #
        # memmap, chunked
        #
        if not silent:
            ConsoleColors.print_green('Testing ' + title + ' with chunked memmap')
        hdr, data_io = BrainVisionReader.bv_read_data(self.bv_data_path, channels=channels,
                                                      start_sample=start_sample, end_sample=end_sample, unit='uV',
                                                      use_memmap=True, chunked_read=True)
        assert_results(hdr, data_io, 'chunked memmap')
        del data_io

        #
        # memmap, non-chunked
        #
        if not silent:
            ConsoleColors.print_green('Testing ' + title + ' with non-chunked memmap')
        hdr, data_io = BrainVisionReader.bv_read_data(self.bv_data_path, channels=channels,
                                                      start_sample=start_sample, end_sample=end_sample, unit='uV',
                                                      use_memmap=True, chunked_read=False)
        assert_results(hdr, data_io, 'non-chunked memmap')
        del data_io


if __name__ == '__main__':
    unittest.main()



"""
Unit tests to validate the EDF (European Data Format) reader

This class tests reading (chunked or non-chunked, standard IO and Memmap):
   - a single channel (all samples)                       --> test01_fileio_edf_validate_data_single_channel
   - the full set (all channels, all samples)             --> test02_fileio_edf_validate_data_fullset
   - 100 000 samples (all channels)                       --> test03_fileio_edf_validate_data_100k_samples
   - multiple (two) ranges of 1000 samples (all channels) --> test04_fileio_edf_validate_data_multi_range

Each of these tests uses the EDF reader to read data and validates the read data against:
    - a specific Matlab data file (.mat) that has been produced by the 'matlab/generate_edf_verification_sets.m' script,
      which in turn uses Fieldtrip to read the data
    - the same subset of data from the full dataset read using MNE


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest
from ieegprep.utils.console import ConsoleColors


class TestFileIOEDFData(unittest.TestCase):
    """
    Validate the EDF (European Data Format) reader output values against the Matlab (fieldtrip) and MNE reader output values
    """

    #
    # Configuration
    #

    matlab_data_path = 'D:\\BIDS_erdetect\\edf_'
    edf_data_path = 'D:\\BIDS_erdetect\\sub-EDF\\ses-ieeg01\\ieeg\\sub-EDF_ses-ieeg01_ieeg.edf'
    #matlab_data_path = os.path.expanduser('~/Documents/ERDetect_perf/edf_')
    #edf_data_path = os.path.expanduser('~/Documents/ERDetect_perf/sub-EDF/ses-ieeg01/ieeg/sub-EDF_ses-ieeg01_ieeg.edf')


    def test01_fileio_edf_validate_data_single_channel(self):
        # validate single channel

        #
        # matlab data
        #
        import scipy.io
        mat = scipy.io.loadmat(self.matlab_data_path + 'allsamples_CH-LMS2.mat')
        self._test_reader_output_matrix(mat['dat'], 0.015, 'Matlab single channel', channels=('LMS2-Ref',))
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_edf
        mne_raw = read_raw_edf(self.edf_data_path, eog=[], misc=[], stim_channel=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here
        try:
            channel_idx = mne_raw.info['ch_names'].index('EEG LMS2-Ref')
        except ValueError:
            self.assertEqual(0, 1, 'Could not find channel \'' + 'LMS2-Ref' + '\' in the dataset')

        self._test_reader_output_matrix(np.array([mne_raw._data[channel_idx, :]]), np.finfo(np.float32).eps, 'MNE single channel', channels=('LMS2-Ref',))
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Single channel test successful')


    def test02_fileio_edf_validate_data_fullset(self):
        # full dataset - all samples - all channels

        # for our test set, ony the channel names without the annotation channel are used. So remove the last
        from ieegprep.fileio.EdfReader import EdfReader
        hdr = EdfReader.edf_read_header(self.edf_data_path)
        channels = hdr['channel_names'][0:-1]

        #
        # matlab data
        #
        import mat73
        mat = mat73.loadmat(self.matlab_data_path + 'allsamples_allchannels.mat')
        self._test_reader_output_matrix(mat['dat'], 0.041, 'Matlab full dataset', channels=channels)
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_edf
        mne_raw = read_raw_edf(self.edf_data_path, eog=[], misc=[], stim_channel=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        self._test_reader_output_matrix(mne_raw._data, np.finfo(np.float32).eps, 'MNE full dataset', channels=channels)
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Full dataset test successful')


    def test03_fileio_edf_validate_data_100k_samples(self):
        # 100 000 samples - all channels

        # for our test set, ony the channel names without the annotation channel are used. So remove the last
        from ieegprep.fileio.EdfReader import EdfReader
        hdr = EdfReader.edf_read_header(self.edf_data_path)
        channels = hdr['channel_names'][0:-1]

        #
        # matlab data
        #
        import scipy.io
        mat = scipy.io.loadmat(self.matlab_data_path + '100ksamples.mat')
        self._test_reader_output_matrix(mat['dat'], 0.041, 'Matlab 100k samples', start_sample=1000, end_sample=101000, channels=channels)
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_edf
        mne_raw = read_raw_edf(self.edf_data_path, eog=[], misc=[], stim_channel=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        self._test_reader_output_matrix(mne_raw._data[:, 1000:101000], np.finfo(np.float32).eps, 'MNE 100k samples', start_sample=1000, end_sample=101000, channels=channels)
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('100k samples test successful')


    """
    def test03b_fileio_edf_validate_data_sliding_ranges(self):
        # sliding ranges - single channel

        # for our test set, ony the channel names without the annotation channel are used. So remove the last
        from ieegprep.fileio.EdfReader import EdfReader
        hdr = EdfReader.edf_read_header(self.edf_data_path)
        channels = hdr['channel_names'][0:-1]

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_edf
        mne_raw = read_raw_edf(self.edf_data_path, eog=[], misc=[], stim_channel=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        # test slide sample over a small number of blocks (1 and 2)
        for start_sample in range(0, 1000, 1):
            for end_sample in range(1000, 400, -1):
                if start_sample < end_sample:
                    print('Test - start sample: ' + str(start_sample) + ' - end sample:' + str(end_sample))
                    self._test_reader_output_matrix(np.array([mne_raw._data[1, start_sample:end_sample]]), np.finfo(np.float32).eps,
                                                    'MNE sliding ranges',
                                                    start_sample=start_sample, end_sample=end_sample, channels=('LAS2-Ref',),
                                                    silent=True)

        # test slide over all sample possibilities where the start is between 0 and 10, and the end is between end-10 and end
        for start_sample in range(0, 10, 1):
            for end_sample in range(hdr['number_of_samples'], hdr['number_of_samples'] - 10, -1):
                print('Test - start sample: ' + str(start_sample) + ' - end sample:' + str(end_sample))
                self._test_reader_output_matrix(mne_raw._data[:, start_sample:end_sample], np.finfo(np.float32).eps,
                                                'MNE sliding ranges',
                                                start_sample=start_sample, end_sample=end_sample, channels=channels,
                                                silent=True)

        # test slide in step of 155 over the sample possibilities where the start is between 0 and 1000, and the end is between end-1000 and end
        for start_sample in range(0, 1000, 155):
            for end_sample in range(hdr['number_of_samples'], hdr['number_of_samples'] - 1000, -155):
                print('Test - start sample: ' + str(start_sample) + ' - end sample:' + str(end_sample))
                self._test_reader_output_matrix(mne_raw._data[:, start_sample:end_sample], np.finfo(np.float32).eps,
                                                'MNE sliding ranges',
                                                start_sample=start_sample, end_sample=end_sample, channels=channels,
                                                silent=True)

        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Sliding ranges test successful')
    """


    def test04_fileio_edf_validate_data_multi_range(self):
        # multiple ranges - all channels

        # for our test set, ony the channel names without the annotation channel are used. So remove the last
        from ieegprep.fileio.EdfReader import EdfReader
        hdr = EdfReader.edf_read_header(self.edf_data_path)
        channels = hdr['channel_names'][0:-1]

        #
        # matlab data
        #
        import scipy.io
        mat = scipy.io.loadmat(self.matlab_data_path + 'multiranges.mat')
        self._test_reader_output_matrix(mat['dat'], 0.041, 'Matlab multiple sample ranges', start_sample=(1000, 101000), end_sample=(2000, 102000), channels=channels)
        del mat

        #
        # MNE data
        #
        import numpy as np
        from mne.io import read_raw_edf
        mne_raw = read_raw_edf(self.edf_data_path, eog=[], misc=[], stim_channel=[], preload=True, verbose=None)
        mne_raw._data *= 1000000    # MNE always returns in V, we dictate mV in our read, so convert MNE data to mV here

        self._test_reader_output_matrix(np.stack((mne_raw._data[:, 1000:2000], mne_raw._data[:, 101000:102000]), axis=-1), np.finfo(np.float32).eps, 'MNE multiple sample ranges', start_sample=(1000, 101000), end_sample=(2000, 102000), channels=channels)
        mne_raw.close()
        del mne_raw._data
        del mne_raw

        ConsoleColors.print_green('Multiple sample ranges test successful')


    def _test_reader_output_matrix(self, mat, diff_tolerance, title, channels=None, start_sample=0, end_sample=-1, silent=False):
        """
        Test the output values of four different types of reading (chunked/non-chucked and stdIO/memmap) against
        a given array of values

        mat (nparray):                      Array of values to test against
        diff_tolerance (float):             The difference between the output values that is tolerated.
                                            The precision of the float values and manipulations from Matlab and the MNE
                                            cause slight differences (~.015/0.041 for matlab, ~.00000000001 for MNE) in
                                            the outputs, this tolerance value allows for less precise testing
        title (str):                        The title of the test

        """
        from ieegprep.fileio.EdfReader import EdfReader
        #import numpy as np

        #
        def assert_results(hdr, data_io, type):

            # find the differences in data and report on the largest difference
            diff = abs(mat - data_io)
            #max_diff = diff.max()
            #if max_diff > 0:
            #    max_diff_index = np.unravel_index(diff.argmax(), diff.shape)
            #    if not silent:
            #        print(' - A maximum difference of ' + str(max_diff) + ' was found at channel ' + hdr['channel_names'][max_diff_index[0]] + ' (index ' + str(max_diff_index[0]) + '). At least at sample: ' + str(max_diff_index[1]) + ' (read value: ' + str(data_io[max_diff_index[0], max_diff_index[1]]) + '; matlab/MNE value: ' + str(mat[max_diff_index[0], max_diff_index[1]]) + ')')

            # assert
            self.assertEqual((diff <= diff_tolerance).sum(), data_io.size, ('Validation of \'' + self.matlab_data_path + '*.mat\' failed while testing ' + title + ' and reading with ' + type))

        #
        # stdIO, chunked
        #
        if not silent:
            ConsoleColors.print_green('> Testing ' + title + ' with chunked stdIO')
        hdr, data_io = EdfReader.edf_read_data(self.edf_data_path, channels=channels,
                                               start_sample=start_sample, end_sample=end_sample, unit='uV',
                                               use_memmap=False, chunked_read=True)
        assert_results(hdr, data_io, 'chunked stdIO')
        del hdr, data_io

        #
        # stdIO, non-chunked
        #
        if not silent:
            ConsoleColors.print_green('> Testing ' + title + ' with non-chunked stdIO')
        hdr, data_io = EdfReader.edf_read_data(self.edf_data_path, channels=channels,
                                               start_sample=start_sample, end_sample=end_sample, unit='uV',
                                               use_memmap=False, chunked_read=False)
        assert_results(hdr, data_io, 'non-chunked stdIO')
        del data_io

        #
        # memmap, chunked
        #
        if not silent:
            ConsoleColors.print_green('> Testing ' + title + ' with chunked memmap')
        hdr, data_io = EdfReader.edf_read_data(self.edf_data_path, channels=channels,
                                               start_sample=start_sample, end_sample=end_sample, unit='uV',
                                               use_memmap=True, chunked_read=True)
        assert_results(hdr, data_io, 'chunked memmap')
        del data_io

        #
        # memmap, non-chunked
        #
        if not silent:
            ConsoleColors.print_green('> Testing ' + title + ' with non-chunked memmap')
        hdr, data_io = EdfReader.edf_read_data(self.edf_data_path, channels=channels,
                                               start_sample=start_sample, end_sample=end_sample, unit='uV',
                                               use_memmap=True, chunked_read=False)
        assert_results(hdr, data_io, 'non-chunked memmap')
        del data_io


if __name__ == '__main__':
    unittest.main()



"""
Functions to read European Data Format (EDF) files


=====================================================
Copyright 2022, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

Adapted from Fieldtrip (by Robert Robert Oostenveld) while replicating some additional header logic from the MNE
package (Teon Brooks, Martin Billinger, Nicolas Barascud, Stefan Appelhoff, Joan Massich, Clemens Brunner, Jeroen Van Der Donckt)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
import os
import math
import logging
import numpy as np
from .IeegDataReader import IeegDataReader


class EdfReader(IeegDataReader):

    edf_hdr = None                   # EDF header
    edf_data = None                  # EDF data (only used on preload)

    def __init__(self, data_path, data_preload=False):
        super().__init__(data_path, data_preload)
        self.data_format = 'edf'

        # load header
        try:
            self.edf_hdr = self.edf_read_header(self.data_path)
        except (FileNotFoundError, IOError):
            logging.error('Could not read EDF header')
            raise RuntimeError('Could not read EDF header')

        # retrieve the sample-rate, total number of samples and channel names
        self.sampling_rate = self.edf_hdr['sampling_frequency']
        self.num_samples = self.edf_hdr['number_of_samples']
        self.channel_names, _ = EdfReader.edf_retrieve_all_channel_names(self.edf_hdr)

        # (optionally) preload data
        if self.data_preload:
            try:
                _, self.edf_data = self.edf_read_data(filepath=self.data_path, hdr=self.edf_hdr, unit='uV')
            except (FileNotFoundError, IOError, TypeError, RuntimeError):
                logging.error('Could not read EDF data')
                raise RuntimeError('Could not read EDF data')


    def close(self):
        del self.edf_hdr
        if self.edf_data is not None:
            del self.edf_data

    def retrieve_channel_data(self, channel_name, ensure_own_data=True):
        """
        Retrieve the channel data (mef = numpy data-array, mne = numpy data-view)

        Args:
            channel_name:                   The name of the channel for which to retrieve the data
            ensure_own_data (bool):         Should ensure the return a numpy array has it's own data (is not a view)

        Returns:
            Numpy array with data

        Raises:
            LookupError:                    Raised when the channel name cannot be found
            RuntimeError:                   Raised when unable to retrieve channel data
        """

        # try to find the channel name
        try:
            channel_index = self.edf_hdr['channel_names'].index(channel_name)
        except ValueError:
            raise LookupError('Could not find channel')

        # determine whether the data is preloaded
        if self.edf_data is None:
            # not preloaded

            # load the channel data
            try:
                _, channel_data = self.edf_read_data(filepath=self.data_path, hdr=self.edf_hdr,
                                                     channels=(channel_name,), unit='uV')
            except Exception:
                logging.error('Could not read EDF data')
                raise RuntimeError('Could not read EDF data')

            return channel_data.squeeze()

        else:
            # preloaded data

            # return the channel data
            if ensure_own_data:
                return self.edf_data[channel_index, :].copy()
            else:
                return self.edf_data[channel_index, :]


    def retrieve_sample_range_data(self, channels, sample_start, sample_end, ensure_own_data=True):
        """
        Retrieve a specific range of EDF data for the requested channels

        Args:
            channels (str, list or tuple):  The channel(s) for which to retrieve the data.
                                            If empty, all channels will be retrieved
            sample_start (int):             The start-point in time (in samples) to start reading from (0-based)
            sample_end (int):               The sample to end the reading at (0-based)
            ensure_own_data (bool):         Should ensure the return a numpy array has it's own data (is not a view)

        Returns:
            data                            A float data matrix containing the signal data (of the requested channels)
                                            formatted as <channels> x <samples/time>

        Raises:
            LookupError:                    Raised when a channel name cannot be found
            RuntimeError:                   Raised when unable to retrieve data
        """

        # check/prepare the channels argument
        if isinstance(channels, str):
            channels = [channels]
        if isinstance(channels, tuple):
            channels = list(channels)
        if channels is None or len(channels) == 0:
            channels = self.channel_names

        if self.edf_data is None:
            # data is not preloaded

            # load the data
            try:
                _, np_sample_data = self.edf_read_data(filepath=self.data_path, hdr=self.edf_hdr,
                                                       channels=channels,
                                                       start_sample=sample_start, end_sample=sample_end)
            except Exception:
                logging.error('Could not read EDF data')
                raise RuntimeError('Could not read EDF data')

            # return the data (from 2-d numpy array to list of numpy arrays)
            return list(np_sample_data)

        else:
            # data is preloaded

            # create a list with the numpy arrays
            sample_data = [None] * len(channels)

            # loop over the requested channels
            for counter in range(len(channels)):

                # find the channel index
                try:
                    channel_index = self.edf_hdr['channel_names'].index(channels[counter])
                except ValueError:
                    raise LookupError('Could not find channel')

                # pick the slice
                if ensure_own_data:
                    sample_data[counter] = self.edf_data[channel_index, sample_start:sample_end].copy()
                else:
                    sample_data[counter] = self.edf_data[channel_index, sample_start:sample_end]

        return sample_data



    #
    # EDF header and data reading functions
    #


    @staticmethod
    def edf_read_header(filepath):
        """
        Read the header of a EDF (.edf) file

        Args:
            filepath (str):                 The path to a EDF file

        Returns:
            A dictionary with header information

        Raises:
            FileNotFoundError:              Raised when unable to find the file
            IOError:                        Raised on any error that occurs while reading or parsing the header

        Examples:
            hdr = edf_read_header('~/dataset.edf')
        """

        #
        # Read/check the file
        #

        # check file existence
        if not os.path.exists(filepath):
            logging.error('EDF file \'' + filepath + '\' could not be found')
            raise FileNotFoundError('No such file or directory: \'' + filepath + '\'')

        # check extension
        if os.path.splitext(filepath)[1].lower() not in ('.edf',):
            logging.error('The file \'' + filepath + '\' has an invalid extension (should be .edf)')
            raise IOError('File \'' + filepath + '\' has an invalid extension')


        try:
            with open(filepath, 'rb') as file:

                hdr = dict()

                #
                # general information
                #

                # read the first part of the header
                header_1 = file.read(256).decode('ascii')

                #
                hdr['version'] = header_1[0:8].rstrip()
                hdr['patient_id'] = header_1[8:8 + 80].rstrip()
                hdr['recording_id'] = header_1[88:88 + 80].rstrip()

                #
                hdr['t0_time'] = list((int(header_1[176:178]), int(header_1[179:181]), int(header_1[182:184])))
                t0_date = list((int(header_1[174:176]), int(header_1[171:173]), int(header_1[168:170])))
                t0_date[0] = t0_date[0] + 2000 if t0_date[0] < 85 else t0_date[0] + 1900
                hdr['t0_date'] = t0_date

                #
                hdr['header_length'] = int(header_1[184:191])
                hdr['num_records'] = int(header_1[236:243])             # number of data records (-1 if unknown)
                hdr['record_duration'] = float(header_1[244:251])       # duration of a data record, in seconds
                hdr['number_of_channels']  = int(header_1[252:255])     # number of channels/signals in data record


                #
                # channels info
                #

                hdr['channel_names'] = []
                hdr['channel_other_info'] = []
                for count in range(0, hdr['number_of_channels']):

                    # add the channel name, and a dict for each channel (to hold all the other information)
                    dict_channel_other_info = dict()
                    channel_info = file.read(16).decode('ascii').rstrip().split(' ')
                    if len(channel_info) > 1:
                        hdr['channel_names'].append(channel_info[1])
                        dict_channel_other_info['type'] = channel_info[0]
                    else:
                        hdr['channel_names'].append(channel_info[0])
                        dict_channel_other_info['type'] = ''
                    hdr['channel_other_info'].append(dict_channel_other_info)

                # transducer
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['transducer'] = file.read(80).decode('ascii').rstrip()

                # physical dim
                for count in range(0, hdr['number_of_channels']):
                    phys_dim = file.read(8).decode('ascii').rstrip()
                    hdr['channel_other_info'][count]['phys_dim'] = phys_dim

                    # determine the unit for specific phys_dim codes (otherwise ensure multiplication by 1)
                    # Note: setting this unit is used to later allow us to convert to the requested output (uV, mV or V)
                    if phys_dim.lower() in ('\u03BCv', '\u00B5v', '\x83\xCAv', 'uv'):
                        hdr['channel_other_info'][count]['units_to_V_gain'] = 1e-6
                    elif phys_dim.lower() == 'mv':
                        hdr['channel_other_info'][count]['units_to_V_gain'] = 1e-3
                    elif phys_dim.lower() == 'v':
                        hdr['channel_other_info'][count]['units_to_V_gain'] = 1
                    else:
                        #logging.warning('The phys_dim on channel ' + hdr['channel_names'][count] + ' is unknown, assume multiplication by 1')
                        hdr['channel_other_info'][count]['units_to_V_gain'] = 1

                # physical min, max, range
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['phys_min'] = np.float64(file.read(8).decode('ascii').rstrip())

                bad_phys_channel_names = []
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['phys_max'] = np.float64(file.read(8).decode('ascii').rstrip())
                    if hdr['channel_other_info'][count]['phys_min'] >= hdr['channel_other_info'][count]['phys_max']:
                        logging.warning('The physical minimum on channel ' + hdr['channel_names'][count] + ' is larger than maximum. Recheck the scaling and polarity')

                    # calculate and check the range
                    hdr['channel_other_info'][count]['phys_range'] = hdr['channel_other_info'][count]['phys_max'] - hdr['channel_other_info'][count]['phys_min']
                    if ~np.isfinite(hdr['channel_other_info'][count]['phys_range']) or hdr['channel_other_info'][count]['phys_range'] == 0:
                        bad_phys_channel_names.append(hdr['channel_names'][count])
                        hdr['channel_other_info'][count]['phys_range'] = 1
                if bad_phys_channel_names:
                    logging.warning('No valid phys scaling factor in channel: ' + ', '.join(bad_phys_channel_names) + ', scaling set to 1 for now')


                # digital min, max, range
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['dig_min'] = np.float64(file.read(8).decode('ascii').rstrip())

                bad_dig_channel_names = []
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['dig_max'] = np.float64(file.read(8).decode('ascii').rstrip())
                    if hdr['channel_other_info'][count]['dig_min'] >= hdr['channel_other_info'][count]['dig_max']:
                        logging.warning('The digital minimum on channel ' + hdr['channel_names'][count] + ' is larger than maximum. Recheck the scaling and polarity')

                    # calculate and check the range
                    hdr['channel_other_info'][count]['dig_range'] = hdr['channel_other_info'][count]['dig_max'] - hdr['channel_other_info'][count]['dig_min']
                    if ~np.isfinite(hdr['channel_other_info'][count]['dig_range']) or hdr['channel_other_info'][count]['dig_range'] == 0:
                        bad_dig_channel_names.append(hdr['channel_names'][count])
                        hdr['channel_other_info'][count]['dig_range'] = 1
                if bad_dig_channel_names:
                    logging.warning('No valid digital scaling factor in channel: ' + ', '.join(bad_dig_channel_names) + ', scaling set to 1 for now')


                # determine scaling and offset
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['cal'] = hdr['channel_other_info'][count]['phys_range'] / hdr['channel_other_info'][count]['dig_range']
                    hdr['channel_other_info'][count]['offset'] = hdr['channel_other_info'][count]['phys_min'] - hdr['channel_other_info'][count]['dig_min'] * hdr['channel_other_info'][count]['cal']

                # prefilters (HighPass, LowPass or Notch)
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['pre_filter'] = file.read(80).decode('ascii').rstrip()
                    # TODO: could split further ("HP:0.1Hz LP:75Hz N:50Hz")

                channel_SPR_equal_tofirst = []
                recordblock_length_in_samples = 0      # the length of a record block in samples (= total of all number of samples-per-record over all channels)
                for count in range(0, hdr['number_of_channels']):
                    hdr['channel_other_info'][count]['samples_per_record'] = int(file.read(8).decode('ascii').rstrip())

                    # count to the total of all number of samples-per-record over all channels
                    recordblock_length_in_samples += hdr['channel_other_info'][count]['samples_per_record']

                    # calculate the sampling rate per channel (depends on the duration of a record)
                    hdr['channel_other_info'][count]['sample_frequency'] = hdr['channel_other_info'][count]['samples_per_record'] / hdr['record_duration']

                    # check if the channel's samples_per_record are equal to the first channel
                    channel_SPR_equal_tofirst.append(hdr['channel_other_info'][0]['samples_per_record'] == hdr['channel_other_info'][count]['samples_per_record'])

                # store the record-block length in the header
                hdr['recordblock_length_in_samples'] = recordblock_length_in_samples


                #
                # checks on header fields based on file size
                #

                # check if the header length matches
                file.read(32 * hdr['number_of_channels'])      # skip
                if file.tell() != hdr['header_length']:
                    logging.error('The position at the end of the header and length of header according to the header do not match')
                    raise IOError('End of the header does not correspond with expected length of the header')

                # calculate and check number of records in the data part of the file
                #   1. determine the total number of bytes in the file
                #   2. number of data bytes = (num_bytes - hdr['header_length'])
                #   3. total number of samples = (num_data_bytes // 2)         # (each sample is 2 bytes for a EDF file)
                #   4. number of records = total number of samples // recordblock length (sum of the samples-per-record over all channels)
                num_bytes = file.seek(0, os.SEEK_END)
                num_records = ((num_bytes - hdr['header_length']) // 2) // recordblock_length_in_samples
                if hdr['num_records'] != num_records:
                    logging.warning('The number of records based on the file size does not match the number of records according to the header. Inferring from file-size')

                    # update num records in header
                    hdr['num_records'] = num_records

                #
                if all(channel_SPR_equal_tofirst):
                    # EDF format (without last annotation channel)

                    hdr['sampling_frequency'] = hdr['channel_other_info'][0]['sample_frequency']
                    hdr['number_of_samples'] = hdr['num_records'] * hdr['channel_other_info'][0]['samples_per_record']


                elif channel_SPR_equal_tofirst[0:-1]:
                    # EDF+ format (last channel are annotations)

                    hdr['sampling_frequency'] = hdr['channel_other_info'][0]['sample_frequency']
                    hdr['number_of_samples'] = hdr['num_records'] * hdr['channel_other_info'][0]['samples_per_record']

                else:
                    logging.warning('Multiple channels have different sampling rates, unable determine a global sampling rate for the header')

            # success, return header dict
            return hdr

        except:
            logging.error('Error while reading the EDF file \'' + filepath + '\'')
            raise IOError('Error reading .edf file')


    @staticmethod
    def edf_retrieve_all_channel_names(hdr):
        """
        Retrieve all of the channel names, while leaving out the annotation channel

        Args:
            hdr (dict):                     The header dictionary as provided by 'edf_read_header'.

        Returns:
            channel_names (list):           A list with all the channel names
            channel_indices (list):         A list with all corresponding channel indices

        Raises:
            RuntimeError:                   Raised upon a processing error
        """

        # determine whether the samples_per_record per channel differ from the first channel
        channel_SPR_equal_tofirst = []
        for count in range(0, hdr['number_of_channels']):
            channel_SPR_equal_tofirst.append(hdr['channel_other_info'][0]['samples_per_record'] == hdr['channel_other_info'][count]['samples_per_record'])

        # try to include all channels
        if all(channel_SPR_equal_tofirst):
            # EDF format (without last annotation channel)

            # all channels
            channel_names = hdr['channel_names']
            channel_indices = list(range(0, len(channel_names)))

        elif channel_SPR_equal_tofirst[0:-1]:
            # EDF+ format (last channel are annotations)

            # all channels except for the last
            channel_names = hdr['channel_names'][0:-1]
            channel_indices = list(range(0, len(channel_names)))

        else:
            logging.error('Multiple channels have different sampling rates, please provide which channels to retrieve (with the same sampling rate')
            raise RuntimeError('Multiple channels have different sampling rates')

        return channel_names, channel_indices


    @staticmethod
    def edf_read_data(filepath, hdr=None, channels=None, start_sample=0, end_sample=-1, unit='uV', use_memmap=True, chunked_read=True):
        """
        Read data from a European Data Format (.edf) file

        Args:
            filepath (str):                 The path to the EDF file
            hdr (dict):                     Optionally pass the header dictionary as provided by 'edf_read_header'. If this
                                            argument is not set, the header file will be parsed before retrieving the data.
                                            If set, this function will use the header that is provided (to prevent multiple
                                            header reads on multiple calls to this function)
            channels (str/list/tuple):      The names of the channels to return the signal data from. The order of channels
                                            in this input argument will determine the order of channels in the output matrix.
                                            If set to None (default), all channels will be read and ordered according to the
                                            header file (hdr['channel_names'])
            start_sample (int or list):     The start-point in time (in samples) to start reading from (0-based). This argument
                                            can either be a single value to indicate the start value of a single range that
                                            should be retrieved over all the requested channels; or this argument can be a
                                            list of start values to retrieve multiple ranges (start- and end-points) for all
                                            the requested channels. When retrieving multiple ranges, make sure the number of
                                            values in the start_sample and the end_sample arguments match, and the size of the
                                            ranges (in samples) are equal.
            end_sample (int or list):       The sample to end the reading (0-based). This argument can either be a single
                                            value to indicate the end-sample of a single range that should be retrieved
                                            over all the requested channels; or this argument can be a list of end values
                                            to retrieve multiple ranges (start- and end-points) for all the requested
                                            channels. When retrieving multiple ranges, make sure the number of values in
                                            the start_sample and the end_sample arguments match, and the size of the ranges
                                            (in samples) are equal. Also note that because the range is 0-based, data are
                                            loaded "up-till" the range end-index, so the result does not include the value
                                            at the end-index (e.g. a requested sample range of 0-3 will return the first 3 values,
                                            being the values at [0], [1], [2]). A value of -1 (default) or -1 value in the list
                                            of values represents the latest sample of the time-series (as indicated by the header)
            unit (str):                     The unit in which the data should be return ('uV' returns microVolts, 'mV' returns
                                            microVolts and 'V' returns Volts)
            use_memmap (bool):              Whether to use numpy's memmap (which wraps around mmap) while reading the data.
                                            If true, the data file is first loaded/cached into virtual memory (pagefile) to
                                            speed up (repetitive) reading. If false, the standard system read operations
                                            will be used (slightly slower for repetitive reading from the file but does not
                                            explicitly require virtual memory).

        Returns:
            hdr                             A dictionary with header information
            data                            A float data matrix containing the signal data (of the requested channels). When
                                            a single time-series is requested (one value as start_sample and one value as
                                            end_sample argument) then the matrix will be formatted as <channels> x <samples/time>,
                                            with the first dimension (rows) representing the channels (ordered based on
                                            the 'channels' input argument) and the second dimension (columns) the samples/time.
                                            If multiple ranges are requested (lists of values for the start_sample and end_sample
                                            arguments) then the return format will be <channels> x <samples/time> x <ranges>, so
                                            that the third dimension represents the requested ranges/epochs.

        Raises:
            FileNotFoundError:              Raised when unable to find the file
            IOError:                        Raised on any error that occurs while reading or parsing the file
            TypeError:                      Raised when input argument types are wrong
            RuntimeError:                   Raised upon a processing error

        Examples:
            hdr, data = edf_read_data('~/dataset.edf')
            hdr, data = edf_read_data('~/dataset.edf', units='uV')
            hdr, data = edf_read_data('~/dataset.edf', channels='CH01')
            hdr, data = edf_read_data('~/dataset.edf', channels=('CH01', 'CH07'))
            hdr, data = edf_read_data('~/dataset.edf', start_sample=1000, end_sample=2000)
            hdr, data = edf_read_data('~/dataset.edf', start_sample=(1000, 6000), end_sample=(2000, 7000))
        """

        # check unit argument
        if unit.lower() not in ('uv', 'mv', 'v'):
            logging.error('Invalid unit ' + unit + ' to retrieve the data in. Only options are uV, mV or V')
            raise RuntimeError('Invalid unit')

        # check file existence
        if not os.path.exists(filepath):
            logging.error('Data file \'' + filepath + '\' could not be found')
            raise FileNotFoundError('No such file or directory: \'' + filepath + '\'')

        # check whether to load the header
        if hdr is None:

            # read the header information
            try:
                hdr = EdfReader.edf_read_header(filepath)
            except (FileNotFoundError, IOError):
                logging.error('Error while reading/parsing header from file \'' + filepath + '\'')
                raise IOError('Error while reading/parsing header')

        else:
            # TODO: check fields in the header file that is passed?
            pass

        # check whether there are samples
        if hdr['number_of_samples'] == 0:
            logging.error('no samples in the data file according to the header')
            raise RuntimeError('no samples in the data')


        #
        # check/prepare channel input(s)
        #

        if isinstance(channels, str):
            channels = (channels,)

        # retrieve channel
        if channels is None or len(channels) == 0:
            # no/empty channels argument

            try:
                _, channel_indices = EdfReader.edf_retrieve_all_channel_names(hdr)
            except RuntimeError:
                raise RuntimeError('Multiple channels have different sampling rates')

        else:
            # select specific channels

            # retrieve the channel indices
            channel_indices = []
            for channel in channels:
                try:
                    index = hdr['channel_names'].index(channel)
                    channel_indices.append(index)
                except ValueError:
                    logging.error('requested channel \'' + channel + '\' was not found')
                    raise IOError('Requested channel was not found')

            # check if all requested channels have the same sampling rate
            for index in channel_indices:
                if hdr['channel_other_info'][channel_indices[0]]['samples_per_record'] != hdr['channel_other_info'][index]['samples_per_record']:
                    logging.error('One or more requested channels differ in sampling rate, can only combine channels with the same sampling rate')
                    raise RuntimeError('Multiple channels have different sampling rates')

        # determine the offsets of each channel in a record
        channel_offsets = []
        record_sample_count = 0
        for count in range(0, hdr['number_of_channels']):
            channel_offsets.append(record_sample_count)
            record_sample_count += hdr['channel_other_info'][count]['samples_per_record']
        del record_sample_count


        #
        # check/prepare sample range input(s)
        #
        single_range, output_dimensions, ranges_start, ranges_length = IeegDataReader.check_sample_arguments(start_sample, end_sample,
                                                                                                             hdr['number_of_samples'],
                                                                                                             len(channel_indices))


        #
        # Read the data file
        #

        # check if the system is 32-bit and memmap is enabled
        # (mapping on 32-bit systems is limited to 2GB files, so disable just in case)
        if use_memmap and not (sys.maxsize > 2 ** 32):
            logging.warning('This is a 32-bit system. Data will be read with standard IO operations (rather than memmap)')
            use_memmap = False

        # helper functions to speed up performance, operation is not performed if it has no effect
        def data_add(in_data, gain):
            return in_data if gain == 0 else in_data + gain

        def data_multiply(in_data, factor):
            return in_data if factor == 1 else in_data * factor

        # determine the unit gains for the requested channels
        channel_gains = []
        for channel_count, channel_index in enumerate(channel_indices):

            unit_gain = hdr['channel_other_info'][channel_index]['units_to_V_gain']
            if unit.lower() == 'uv':
                unit_gain *= 1e+6
            elif unit.lower() == 'mv':
                unit_gain *= 1e+3

            channel_gains.append(unit_gain)
        del channel_count, channel_index

        # quick ref for SPR
        samples_per_record = hdr['channel_other_info'][channel_indices[0]]['samples_per_record']

        # determine the sample length
        range_length = output_dimensions[1]

        # int16 data could be cast to float32 (using less memory than a float64) but EDF will very often
        # have an offset, cal or unit gain value. So default to float64 to accommodate more precision
        data = np.empty(output_dimensions, dtype=np.float64)

        # determine whether to read chunked
        #if use_memmap:
        #    chunked_read = True
        #else:
            # Note: the consideration here depends on how the number of channels to be retrieved compares to the
            #       number of channels (in a record), see tipping point 4/149 and 10/149 channel benchmark.
            #       More testing is needed. However, often just a single channel is loaded, in this case non-chunked
            #       is faster; else wise chunked read.
        #    chunked_read = len(channel_indices) != 1

        #
        if chunked_read:
            # chunked reading
            #print('chunked - memmap = ' + str(use_memmap))

            if use_memmap:
                amem = np.memmap(filepath, dtype="<i2", mode='r', offset=hdr['header_length'], order='C')
            else:
                try:
                    fid_data = open(filepath, "rb")
                except IOError:
                    logging.error('Error while opening data file \'' + filepath + '\'')
                    raise IOError('Error while opening data file')

            # determine how many record there are in each chunk (samplesize is 2 bytes for EDF), assuming chunks of ~10mb
            # rounded down to the number of channels
            record_chunk_size = ((int(10e6) // 2) // hdr['recordblock_length_in_samples'])
            if record_chunk_size < 1:
                record_chunk_size = 1

            # define read function for reading binary in chunks
            def read_binary_chunked(rb_range_index, rb_start, rb_length):

                # determine
                begin_record = math.floor(rb_start / samples_per_record)

                # determine the record offset in samples
                record_offset_in_samples = begin_record * hdr['recordblock_length_in_samples']

                # edf = <dim0=records> x <dim1=channels> x <dim2=samples>
                data_sample_pos = 0
                while data_sample_pos < rb_length:
                    samples_left = rb_length - data_sample_pos

                    # determine the number of records need to be read for the current chunk
                    #
                    # Note:  In the situation where: (1) the first read occurs and the requested number of samples (rb_length)
                    #        results in requiring only one chunked read (less records needed than the maximum); and (2) the
                    #        requested start_sample is not at the start of a record (start % samples_per_record); then
                    #        the chunk might require an extra record because the offset shift.
                    # Note2: The chunks after the first read will always start at the beginning of the record (no offset) so
                    #        an extra record will never be needed
                    if data_sample_pos == 0:
                        current_chunk_num_records = min(math.ceil((rb_start % samples_per_record + samples_left) / samples_per_record), record_chunk_size)
                    else:
                        current_chunk_num_records = min(math.ceil(samples_left / samples_per_record), record_chunk_size)

                    # read the data and reshape on records
                    if use_memmap:
                        chunk_data = amem[record_offset_in_samples:record_offset_in_samples + current_chunk_num_records * hdr['recordblock_length_in_samples']].reshape(current_chunk_num_records, -1, order='C')
                    else:
                        fid_data.seek(hdr['header_length'] + record_offset_in_samples * 2, os.SEEK_SET)
                        chunk_data = np.fromfile(fid_data, dtype="<i2", count=current_chunk_num_records * hdr['recordblock_length_in_samples']).reshape(current_chunk_num_records, -1, order='C')

                    # determine where in the chunk where to start and where to end
                    if data_sample_pos == 0:
                        current_chunk_begin_sample = rb_start - begin_record * samples_per_record
                    else:
                        current_chunk_begin_sample = 0
                    current_chunk_end_sample = min(current_chunk_begin_sample + samples_left, samples_per_record * current_chunk_num_records)
                    num_samples = current_chunk_end_sample - current_chunk_begin_sample

                    # loop over the channels
                    for rb_channel_count, rb_channel_index in enumerate(channel_indices):
                        channel_data = chunk_data[:, channel_offsets[rb_channel_index]:channel_offsets[rb_channel_index] + samples_per_record].flatten()

                        if rb_range_index is None:
                            data[rb_channel_count, data_sample_pos:data_sample_pos + num_samples] = data_multiply(data_add(data_multiply(channel_data[current_chunk_begin_sample:current_chunk_end_sample], hdr['channel_other_info'][rb_channel_index]['cal']), hdr['channel_other_info'][rb_channel_index]['offset']), channel_gains[rb_channel_count])
                            #data[rb_channel_count, data_sample_pos:data_sample_pos + num_samples] = channel_data[record_begin_sample:record_end_sample]

                        else:
                            data[rb_channel_count, data_sample_pos:data_sample_pos + num_samples, rb_range_index] = data_multiply(data_add(data_multiply(channel_data[current_chunk_begin_sample:current_chunk_end_sample], hdr['channel_other_info'][rb_channel_index]['cal']), hdr['channel_other_info'][rb_channel_index]['offset']), channel_gains[rb_channel_count])
                            #data[rb_channel_count, data_sample_pos:data_sample_pos + num_samples, rb_range_index] = channel_data[record_begin_sample:record_end_sample]

                    # move to the next chunk
                    data_sample_pos += num_samples
                    record_offset_in_samples += current_chunk_num_records * hdr['recordblock_length_in_samples']

            # read range(s)
            if single_range:
                read_binary_chunked(None, ranges_start, ranges_length)
            else:
                for range_index in range(len(ranges_start)):
                    read_binary_chunked(range_index, ranges_start[range_index], ranges_length[range_index])

            # close file if needed
            if not use_memmap and fid_data is not None:
                fid_data.close()


        else:
            # non-chunked reading
            #print('non-chunked - memmap = ' + str(use_memmap))

            if use_memmap:

                # open memory map
                amem = np.memmap(filepath, dtype="<i2", mode='r', offset=hdr['header_length'], shape=(hdr['num_records'], hdr['recordblock_length_in_samples']), order='C')

                # read function
                def read_range(rb_range_index, rb_start, rb_end):

                    # determine which records would hold this range of samples
                    begin_record = math.floor(rb_start / samples_per_record)
                    end_record = math.floor(rb_end / samples_per_record)

                    sample_offset = rb_start - begin_record * samples_per_record

                    # loop over the channels
                    for rb_channel_count, rb_channel_index in enumerate(channel_indices):

                        # pick channel data and flatten
                        channel_data = np.array(amem[begin_record:end_record + 1, channel_offsets[rb_channel_index]:channel_offsets[rb_channel_index] + hdr['channel_other_info'][rb_channel_index]['samples_per_record']]).flatten()

                        # retrieve the data (with cal, offset and gain applied)
                        if rb_range_index is None:
                            data[rb_channel_count, :] = data_multiply(data_add(data_multiply(channel_data[sample_offset:sample_offset + range_length], hdr['channel_other_info'][rb_channel_index]['cal']), hdr['channel_other_info'][rb_channel_index]['offset']), channel_gains[rb_channel_count])
                            #data[rb_channel_count, :] = channel_data[sample_offset:sample_offset + range_length]
                        else:
                            data[rb_channel_count, :, rb_range_index] = data_multiply(data_add(data_multiply(channel_data[sample_offset:sample_offset + range_length], hdr['channel_other_info'][rb_channel_index]['cal']), hdr['channel_other_info'][rb_channel_index]['offset']), channel_gains[rb_channel_count])
                            #data[rb_channel_count, :, rb_range_index] = channel_data[sample_offset:sample_offset + range_length]


            else:
                # frombuffer (instead of memmap)

                # open file
                try:
                    fid_data = open(filepath, "rb")
                except IOError:
                    logging.error('Error while opening data file \'' + filepath + '\'')
                    raise IOError('Error while opening data file')

                # read function
                def read_range(rb_range_index, rb_start, rb_end):

                    # determine which records would hold this range of samples
                    begin_record = math.floor(rb_start / samples_per_record)
                    end_record = math.floor(rb_end / samples_per_record)

                    sample_offset = rb_start - begin_record * samples_per_record

                    # first loop over records, then channels (memmap only),
                    data_index = 0
                    for record_index in range(begin_record, end_record + 1):

                        #
                        if record_index == begin_record:
                            record_begin_sample = sample_offset
                        else:
                            record_begin_sample = 0

                        if record_index == end_record:
                            record_end_sample = range_length - data_index + record_begin_sample
                        else:
                            record_end_sample = samples_per_record

                        #
                        data_length = record_end_sample - record_begin_sample
                        record_offset = hdr['header_length'] + record_index * hdr['recordblock_length_in_samples'] * 2

                        for rb_channel_count, rb_channel_index in enumerate(channel_indices):

                            channel_begin_sample = record_begin_sample + channel_offsets[rb_channel_index]
                            channel_num_sample = record_end_sample - record_begin_sample

                            fid_data.seek(record_offset + channel_begin_sample * 2, os.SEEK_SET)
                            if rb_range_index is None:
                                data[rb_channel_count, data_index:data_index + data_length] = data_multiply(data_add(data_multiply(np.frombuffer(fid_data.read(channel_num_sample * 2), dtype='<i2'), hdr['channel_other_info'][rb_channel_index]['cal']), hdr['channel_other_info'][rb_channel_index]['offset']), channel_gains[rb_channel_count])
                                #data[rb_channel_count, data_index:data_index + data_length] = np.frombuffer(fid_data.read(channel_num_sample * 2), dtype='<i2')
                            else:
                                data[rb_channel_count, data_index:data_index + data_length, rb_range_index] = data_multiply(data_add(data_multiply(np.frombuffer(fid_data.read(channel_num_sample * 2), dtype='<i2'), hdr['channel_other_info'][rb_channel_index]['cal']), hdr['channel_other_info'][rb_channel_index]['offset']), channel_gains[rb_channel_count])
                                #data[rb_channel_count, data_index:data_index + data_length, rb_range_index] = fid_data.read(channel_num_sample * 2), dtype='<i2')

                        #
                        data_index += data_length


            # read the range(s)
            if single_range:
                read_range(None, ranges_start, ranges_start + ranges_length)
            else:
                for range_index in range(len(ranges_start)):
                    read_range(range_index, ranges_start[range_index], ranges_start[range_index] + ranges_length[range_index])

            # close file if needed
            if not use_memmap and fid_data is not None:
                fid_data.close()


                """
                # test code for first loop over channels, then records

                # determine which records would hold this range of samples
                begin_record = math.floor(read_start_sample / samples_per_record)
                end_record = math.floor(read_end_sample / samples_per_record)

                sample_offset = read_start_sample - begin_record * samples_per_record

                # test code for first loop over channels, then records (memmap & frombuffer), single range only (because writes directly to data)
                fid_data = open(filepath, "rb")  # slower (but still an option if memmap is not supported or slower)
                for rb_channel_count, rb_channel_index in enumerate(channel_indices):

                    data_index = 0
                    for record_index in range(begin_record, end_record + 1):

                        if record_index == begin_record:
                            record_begin_sample = sample_offset
                        else:
                            record_begin_sample = 0

                        if record_index == end_record:
                            record_end_sample = range_length - data_index + record_begin_sample
                        else:
                            record_end_sample = samples_per_record

                        data_length = record_end_sample - record_begin_sample

                        record_begin_sample += channel_offsets[rb_channel_index]
                        record_end_sample += channel_offsets[rb_channel_index]

                        fid_data.seek(hdr['header_length'] + (record_index * hdr['recordblock_length_in_samples'] + record_begin_sample) * 2, os.SEEK_SET)
                        data[rb_channel_count, data_index:data_index + data_length] = np.frombuffer(fid_data.read(data_length * 2), dtype='<i2')
                        #data[rb_channel_count, data_index:data_index + data_length] = amem[record_index, record_begin_sample:record_end_sample]
                        data_index += data_length

                """

        # return the header information and the data
        return hdr, data



# define directory as package

# flatten access
from ieegprep.bids.data_structure import list_bids_datasets
from ieegprep.bids.sidecars import load_channel_info, load_event_info, load_ieeg_sidecar
#from data_epoch import load_data_epochs, load_data_epochs_averages




"""
Functions to search and list BIDS datasets


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import re
import posixpath
import fnmatch
import logging
from ieegprep.fileio import VALID_FORMAT_EXTENSIONS


def list_bids_datasets(bids_search_directory, dataset_extensions=None, subjects_filter=None, subset_search_pattern=None, strict_search=False, only_subjects_with_subsets=False):
    """
    Search and list the subjects and datasets for a given BIDS directory. The BIDS directory functions as a search
    starting point and can point to  any depth of the BIDS structure. Only datasets that are found (recursively) from
    that search depth are included.

    Args:
        bids_search_directory (str):        The BIDS directory to search subject and datasets in
        dataset_extensions (list/tuple):    The file extensions that will be considered as datasets. If set to None (default)
                                            all the readable dataset formats will be considered (EDF, BrainVision, MEF3)
        subjects_filter (list/tuple):       Optional list of subjects to include. If set, then any subject found that is
                                            not included in this list will be excluded from the results. Pass
                                            None (default) to include all found subjects.
        subset_search_pattern (str/list):   Optional search string or tuple/list of search strings. If a dataset name
                                            contains any of these subset patterns, then it will be included in the search
                                            results. Set to None (default) to include all datasets.
        strict_search (bool):               List only subjects and datasets that abide (more) strictly to the BIDS folder
                                            and filename structure, such as whether to require a dataset to be in a modality
                                            folder and matching of the subject-label between dataset names and their directories.
                                            Set to True to be more strict, False (default) for less strict searching.
        only_subjects_with_subsets (bool):  Whether to list a subject in the results that has no listed datasets after
                                            applying the search arguments (dataset_extensions, subjects_filter and
                                            subset_search_pattern) in it. Set to True in order to only add subjects to
                                            the results that have any datasets. False (default) to also list subjects
                                            without datasets.

    Returns:
        A dictionary with subjects and their corresponding datasets. The keys of the dictionary represent the
        subject-labels. Each dictionary value holds a list with paths to datasets for that specific subject.

    Raises:
        NotADirectoryError:                 If the BIDS input path does not exist or is not a directory
        ValueError:                         Invalid subset_search_pattern or dataset_extensions parameter
    """

    # retrieve the absolute/resolved path
    bids_search_directory = os.path.abspath(os.path.expanduser(os.path.expandvars(bids_search_directory)))
    if not os.path.isdir(bids_search_directory):
        logging.error('The BIDS input path \'' + bids_search_directory + '\' does not exists or is not a directory')
        raise NotADirectoryError('The BIDS input path does not exists or is not a directory')

    # prepare the search patterns parameter
    if subset_search_pattern is None or not subset_search_pattern:
        subset_patterns = ('',)
    elif isinstance(subset_search_pattern, str):
        subset_patterns = (subset_search_pattern.lower(), )
    elif isinstance(subset_search_pattern, list) or isinstance(subset_search_pattern, tuple):
        subset_patterns = subset_search_pattern
        subset_patterns = list(set(subset_patterns))            # make unique
        subset_patterns = [x.lower() for x in subset_patterns]  # lower-case
    else:
        logging.error('Invalid subset_search_pattern parameter, this argument can either be None (default), a single search string or a tuple/list of strings')
        raise ValueError('Invalid subset_search_pattern parameter')

    # prepare the data formats parameter
    if dataset_extensions is not None and dataset_extensions:
        extensions = dataset_extensions
        for extension in extensions:
            if not any(extension in x for x in VALID_FORMAT_EXTENSIONS):
                logging.error('Invalid dataset_extensions parameter, this argument can either be None (default), or a tuple/list of combining one or more valid extensions (valid extension are: ' + ', '.join(VALID_FORMAT_EXTENSIONS) + ').')
                raise ValueError('Invalid dataset_extensions parameter')
    else:
        extensions = VALID_FORMAT_EXTENSIONS

    # prepare subject filter parameter
    if not subjects_filter is None and subjects_filter:

        # if subject should be filtered, then lower-case the filter-subjects and remove the 'sub-' prefix
        subjects_filter = [s.lower().replace('sub-', '') for s in subjects_filter]


    #
    # inventorize the subject(s)
    #

    datasets = dict()
    subject_dirs = dict()

    # check if there as subject ('sub-') directories in the input directory
    #
    # Note:   First check root, and only then single directory. This will ensure that even something unlikely
    #         as '/desktop/sub-test/BIDS_root' will work as an input path
    # Note 2: next/os.walk is a fast method to find subject folders, see test_list_subject_dirs_perf.py
    sub_folders = next(os.walk(bids_search_directory))[1]
    sub_folders = [f for f in sub_folders if f.lower().startswith('sub-')]
    if sub_folders:
        # the input directory is in a BIDS root

        # apply subject filter if needed
        if not subjects_filter is None and subjects_filter:

            # filter and add each subject as a subject directory (with the subject folder for further searching of subsets)
            for subject_label in sub_folders:
                if subject_label[4:].lower() in subjects_filter:
                    subject_dirs[subject_label] = os.path.join(bids_search_directory, subject_label)
        else:

            # add each subject as a subject directory (with the subject folder for further searching of subsets)
            for subject_label in sub_folders:
                subject_dirs[subject_label] = os.path.join(bids_search_directory, subject_label)

    else:
        # perhaps the input path is in a subject (sub)directory (or a directory with data files where we can extract the subject name from)

        # subtract the name (returns without 'sub-' prefix)
        subject_label = _extract_subject_label_from_path(bids_search_directory)
        if subject_label is None:
            # no subject name could be derived from the input path, this already breaks the BIDS specification.

            if strict_search:
                print('Warning: The input directory is neither a BIDS root directory (no subject directories were found)\n'
                      '         nor a BIDS sub-directory (unable to extract a \'sub-<label>\' from the BIDS input\n'
                      '         path. The BIDS standard dictates that data files should be in a subject directory).')

            else:
                # try to search for datasets further down the path

                print('Warning: The input directory is neither a BIDS root directory (no subject directories were found)\n'
                      '         nor a BIDS sub-directory (unable to extract a \'sub-<label>\' from the BIDS input path. The BIDS\n'
                      '         standard dictates that data files should be in a subject directory).\n\n'
                      '         Now recursively searching the directory for BIDS data files (that according to the BIDS\n'
                      '         standard should include a subject-label in their filename)...')

                # search for files
                data_files = _search_directory_for_datasets(bids_search_directory, extensions, name_search_patterns=subset_patterns, modalities=None)

                # loop over the results and split out the different subjects with their corresponding subjects labels
                for file in data_files:
                    file_subject_label = _extract_subject_label_from_path(file)

                    if file_subject_label is None:
                        print('Warning: Unable to determine the subject label for dataset \'' + file + '\'. Excluded from listing.')
                    else:

                        # apply the subject filter if there is one
                        if not subjects_filter is None and subjects_filter:
                            if not file_subject_label.lower() in subjects_filter:
                                continue

                        # add the subject and datafiles
                        if not file_subject_label in datasets.keys():
                            datasets['sub-' + file_subject_label] = []
                        datasets['sub-' + file_subject_label].append(file)

        else:
            # subject label was found

            # add subject as a subject directory (with the subject folder for further searching of subsets)
            # apply the subject filter if there is one
            if not subjects_filter is None and subjects_filter:
                if subject_label.lower() in subjects_filter:
                    subject_dirs['sub-' + subject_label] = bids_search_directory
            else:
                subject_dirs['sub-' + subject_label] = bids_search_directory


    # loop over the subject directories
    for dir_subject_label, subject_path in subject_dirs.items():

        # search file data files in the subject directory (given more or less constrains)
        if strict_search:
            data_files = _search_directory_for_datasets(subject_path, extensions, name_search_patterns=subset_patterns, modalities=('ieeg', 'eeg'))
        else:
            data_files = _search_directory_for_datasets(subject_path, extensions, name_search_patterns=subset_patterns, modalities=None)

        # loop over the datasets that were found in the subject dir
        for file in data_files:
            file_subject_label = _extract_subject_label_from_path(file)

            # check if the file's subject label corresponds with the subject-label from the directory
            if file_subject_label is None:
                if strict_search:
                    print('Warning: Unable to determine the subject-label for dataset \'' + file + '\'. Excluded file from listing.')
                    continue
                else:
                    print('Warning: Unable to determine the subject-label for dataset \'' + file + '\'.\n'
                          '         Assuming the subject-label from the directory as the subject-label for the file')
                    file_subject_label = dir_subject_label[4:]

            # check if file subject label is equal to the sub-folder subject label
            if not file_subject_label == dir_subject_label[4:]:
                if strict_search:
                    print('Warning: The subject-label in the file (\'sub-' + file_subject_label + '\') and the subject-label\n'
                          '         in the directory (\'' + dir_subject_label + '\') do not match.\n'
                          '         Excluded file from listing.')
                    continue

                else:
                    print('Warning: The subject-label in the file (\'sub-' + file_subject_label + '\') and the subject-label\n'
                          '         in the directory (\'' + dir_subject_label + '\') do not match.\n'
                          '         Assuming the subject-label from the file is correct')

                # it is possible that the subject-label from the file has replaced subject-label from the directory by now, therefore
                # we re-check whether the subject-label (from the file) against the included subjects list here
                if not subjects_filter is None and subjects_filter and not file_subject_label.lower() in subjects_filter:
                    continue

            # add the subject and data file
            if not ('sub-' + file_subject_label) in datasets.keys():
                datasets['sub-' + file_subject_label] = []
            datasets['sub-' + file_subject_label].append(file)

        # empty subjects might also need to be listed
        # if so, check to make sure an empty entry exists (in case no files were added above)
        if not only_subjects_with_subsets and not dir_subject_label in datasets.keys():
            datasets[dir_subject_label] = []


    # sort the subjects and datasets
    dataset_sorted = dict()
    for subject in sorted(datasets):
        dataset_sorted[subject] = datasets[subject]
        dataset_sorted[subject].sort()

    #
    return dataset_sorted


def _extract_subject_label_from_path(path):
    """
    Extract the BIDS subject-label from a given path. This can either be a BIDS directory or a BIDS datafile.
    If the path points to a file (determined by the presence of a file-extension), then the subject-label will be
    extracted from the filename and the directory in which the file resides will be ignored.

    Args:
        path (str):                         The directory or filepath to extract the subject label from

    Returns:
        The subject label (without the 'sub-' prefix) if successful; None on failure to extract

    """

    # detect the prefix
    sub_start = path.rfind('sub-')
    if sub_start != -1:

        # check if file (every BIDS file has an extension)
        sub_end = path.find('.', sub_start)

        if sub_end != -1:
            # assume file

            # check for directory seperator (if file, there should not be a file seperator after the 'sub-' part)
            seperator_pos = max(path.find('/', sub_start), path.find('\\', sub_start))
            if seperator_pos == -1:
                # no directory-seperator, can continue to assume file

                # extract the subject-name from the file
                sub_end = path.find('_', sub_start)
                if sub_end != -1 and sub_end != sub_start + 4:
                    return path[sub_start + 4:sub_end]

        else:
            # assume directory

            sub_end = path.find('_', sub_start)
            if sub_end == -1:                                       sub_end = max(path.find('/', sub_start), path.find('\\', sub_start))
            if sub_end == -1 and len(path[sub_start:]) > 4:         sub_end = len(path)

            if sub_end != -1 and sub_end != sub_start + 4:
                return path[sub_start + 4:sub_end]

    # return failure
    return None


def _search_directory_for_datasets(search_path, dataset_extensions, name_search_patterns=('',), modalities=None):
    """
    Search the input directory recursively for (BIDS) datasets

    Args:
        search_path (str):                  The path to the directory to (recursively) search
        dataset_extensions (list/tuple):    The extensions that will be considered as datasets
        name_search_patterns (list/tuple):  A list of search strings. If a dataset name contains any of these subset
                                            patterns it will be included in the search results. The pattern search is
                                            case-insensitive
        modalities (list/tuple):            A list of BIDS modalities. If specified, the function will only search for
                                            datasets in BIDS paths that match any of the specified modalities (e.g. '*/ieeg/*')
                                            To disable the modalities requirement, set this argument to None.

    Returns:
        A list of datasets that were found in the directory

    """

    # reproduce the filter function from fnmatch but compile a case-insensitive regex
    # Note: this allows the search to be case-insensitive, but will leave the casing intact on the returned filenames (important file unix filepaths)
    # (which would be more complicated in a solution where all the search patterns and all filenames would be lowered)
    def filter_case_insensitive(names, pat):
        result = []
        pat = os.path.normcase(pat)
        regex = fnmatch.translate(pat)
        match = re.compile(regex, re.IGNORECASE).match
        if os.path is posixpath:
            for name in names:
                if match(name):
                    result.append(name)
        else:
            for name in names:
                if match(os.path.normcase(name)):
                    result.append(name)
        return result


    # loop over all folders within the search path
    subsets = []
    for root, dirs, files in os.walk(search_path):
        if modalities is None or root.lower().endswith(modalities):
            for extension in dataset_extensions:
                for search_pattern in name_search_patterns:
                    search_pattern = '*' + search_pattern + '*' if search_pattern else '*'
                    subsets.extend([os.path.join(root, f) for f in filter_case_insensitive(files, search_pattern + extension)])
                    subsets.extend([os.path.join(root, f) for f in filter_case_insensitive(dirs, search_pattern + extension)])

    # in the case of multiple search patterns, filenames can satisfy multiple patterns and therefore occur multiple
    # times in the results. Check and remove duplicates here
    if len(name_search_patterns) > 1:
        subsets = [*set(subsets)]

    # bring subsets with multiple formats/extensions down to one format (prioritized to occurrence in the extension var)
    for subset in subsets:
        subset_name = subset[:subset.rindex(".")]
        for subset_other in reversed(subsets):
            if not subset == subset_other:
                subset_other_name = subset_other[:subset_other.rindex(".")]
                if subset_name == subset_other_name:
                    subsets.remove(subset_other)

    # return the results
    return subsets



"""
Universal IEEG data reader class


=====================================================
Copyright 2022, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from abc import abstractmethod


#
# constants
#
VALID_FORMAT_EXTENSIONS         = ('.mefd', '.edf', '.vhdr', '.vmrk', '.eeg')   # valid data format to search for (European Data Format, BrainVision and MEF3)


class IeegDataReader(object):

    initialized = False
    data_path = ''
    data_preload = False

    data_format = ''                # the type of data ('bv', 'edf' or 'mef3')
    sampling_rate = -1              # the sampling rate of the dataset (from metadata, assumes same rate over channels)
    num_samples = -1                # the total number of samples per channel (from metadata, assumes same rate over channels)
    channel_names = []              # all of the channel names

    def __new__(cls, data_path, data_preload=False):

        # return a sub-class instance depending on the format
        try:
            data_extension = data_path[data_path.rindex("."):]
            if data_extension == '.edf':
                from .EdfReader import EdfReader
                return super(IeegDataReader, cls).__new__(EdfReader)

            elif data_extension == '.vhdr' or data_extension == '.vmrk' or data_extension == '.eeg':
                from .BrainVisionReader import BrainVisionReader
                return super(IeegDataReader, cls).__new__(BrainVisionReader)

            elif data_extension == '.mefd':
                from .Mef3Reader import Mef3Reader
                return super(IeegDataReader, cls).__new__(Mef3Reader)

            else:
                logging.error('Unknown data format (' + data_extension + ')')
                raise ValueError('Unknown data format')

        except ValueError:
                logging.error('Invalid data path (' + data_path + ')')
                raise ValueError('Invalid data path')

    def __init__(self, data_path, data_preload=False):
        self.data_path = data_path
        self.data_preload = data_preload

    @abstractmethod
    def close(self): pass

    @abstractmethod
    def retrieve_channel_data(self, channel_name, ensure_own_data=True): pass

    @abstractmethod
    def retrieve_sample_range_data(self, channels, sample_start, sample_end, ensure_own_data=True): pass

    @staticmethod
    def check_sample_arguments(start_sample, end_sample, number_of_samples, number_of_channels):
        """
        Helper function to check the start_sample and end_sample arguments

        Args:
            start_sample (int or list):     The start-point in time (in samples) to start reading from (0-based). This argument
                                            can either be a single value to indicate the start value of a single range that
                                            should be retrieved over all the requested channels; or this argument can be a
                                            list of start values to retrieve multiple ranges (start- and end-points) for all
                                            the requested channels. When retrieving multiple ranges, make sure the number of
                                            values in the start_sample and the end_sample arguments match, and the size of the
                                            ranges (in samples) are equal.
            end_sample (int or list):       The sample to end the reading (0-based). This argument can either be a single
                                            value to indicate the end-sample of a single range that should be retrieved
                                            over all the requested channels; or this argument can be a list of end values
                                            to retrieve multiple ranges (start- and end-points) for all the requested
                                            channels. When retrieving multiple ranges, make sure the number of values in
                                            the start_sample and the end_sample arguments match, and the size of the ranges
                                            (in samples) are equal. Also note that because the range is 0-based, data are
                                            loaded "up-till" the range end-index, so the result does not include the value
                                            at the end-index (e.g. a requested sample range of 0-3 will return the first 3 values,
                                            being the values at [0], [1], [2]). A value of -1 (default) or -1 value in the list
                                            of values represents the latest sample of the time-series (as indicated by the header)
            number_of_samples (int):        The number of samples (for each channel) in the dataset, is used to check boundaries
            number_of_channels (int):       The number of channels, is only used to output the output matrix dimension in one go

        Returns:
            single_range:                   True if single range, False if multiple ranges
            output_dimensions (tuple):      The dimensions that the output matrix should be based on the ranges
            range_start (int or list):      Int (single range) or list (multiple ranges) containing the start sample(s) of each requested range
            range_length (int or list):     Int (single range) or List (multiple ranges) containing the length(s) of each requested range

        Raises:
            TypeError:                      Raised when input argument types are wrong
            RuntimeError:                   Raised upon a processing error
        """

        def check_input_range(in_start_sample, in_end_sample):

            # fraction
            if isinstance(in_start_sample, float) and not in_start_sample.is_integer():
                logging.error('invalid start_sample argument (' + str(in_start_sample) + '), cannot be a factional value')
                raise TypeError('invalid start_sample argument')
            if isinstance(in_end_sample, float) and not in_end_sample.is_integer():
                logging.error('invalid end_sample argument (' + str(in_end_sample) + '), cannot be a factional value')
                raise TypeError('invalid end_sample argument')

            # start
            if in_start_sample < 0:
                logging.error('invalid start_sample input argument (' + str(in_start_sample) + '), should be 0 or higher')
                raise RuntimeError('invalid start_sample argument')
            if in_start_sample > number_of_samples:
                logging.error('invalid start_sample input argument (' + str(in_start_sample) + '), exceeds the number of sample in the data (' + str(number_of_samples) + ')')
                raise RuntimeError('invalid start_sample argument')

            # end
            if in_end_sample == -1:
                out_end_sample = number_of_samples
            elif in_end_sample < 1:
                logging.error('invalid end_sample input argument (' + str(in_end_sample) + '), should either -1 or >0')
                raise RuntimeError('invalid end_sample argument')
            elif in_end_sample > number_of_samples:
                logging.error('invalid end_sample input argument (' + str(in_end_sample) + '), exceeds the number of sample in the data (' + str(number_of_samples) + ')')
                raise RuntimeError('invalid end_sample argument')
            else:
                out_end_sample = in_end_sample

            # range
            if out_end_sample < in_start_sample:
                logging.error('invalid start_sample and end_sample input arguments, end-point sample (' + str(out_end_sample) + ') should be after the start-point sample (' + str(start_sample) + ')')
                raise RuntimeError('invalid start_sample and end_sample arguments')
            if out_end_sample - in_start_sample == 0:
                logging.error('invalid start_sample (' + str(in_start_sample) + ') and end_sample (' + str(out_end_sample) + ') input arguments, the requested range length is 0')
                raise RuntimeError('invalid start_sample and end_sample arguments')

            return in_start_sample, out_end_sample, out_end_sample - in_start_sample


        # type (specific) checks
        if (isinstance(start_sample, list) or isinstance(start_sample, tuple)) and (isinstance(end_sample, float) or isinstance(end_sample, int)) or \
           (isinstance(end_sample, list) or isinstance(end_sample, tuple)) and (isinstance(start_sample, float) or isinstance(start_sample, int)):
            logging.error('Start_sample and end_sample input argument types do not match. Both arguments should either be single values or lists/tuples of values')
            raise TypeError('Start_sample and end_sample input argument types do not match')

        elif (isinstance(start_sample, float) or isinstance(start_sample, int)) and (isinstance(end_sample, float) or isinstance(end_sample, int)):
            # single value checks

            # flag as single range and check input
            single_range = True
            try:
                range_start, _, range_length = check_input_range(start_sample, end_sample)
            except RuntimeError:
                raise RuntimeError('Error on input arguments')

            # set output dimensions
            output_dimensions = [number_of_channels, range_length]

        elif (isinstance(start_sample, list) or isinstance(start_sample, tuple)) and (isinstance(end_sample, list) or isinstance(end_sample, tuple)):
            # multiple value checks

            # check the lists
            if len(start_sample) != len(end_sample):
                logging.error('Number of values in the start_sample and end_sample arguments do not match. To retrieve multiple ranges/epochs, make sure there are an equal amount of start-points and end-points in the input arguments')
                raise RuntimeError('Number of values in the start_sample and end_sample arguments do not match')
            if len(start_sample) == 0:
                logging.error('The start_sample and end_sample arguments are empty')
                raise RuntimeError('Empty start_sample and end_sample arguments')

            # flag as multiple ranges and check input ranges
            single_range = False
            range_start = []
            range_length = []
            for index in range(len(start_sample)):
                try:
                    i_range_start, _, i_range_length = check_input_range(start_sample[index], end_sample[index])
                    range_start.append(i_range_start)
                    range_length.append(i_range_length)
                    del i_range_start, i_range_length

                except RuntimeError:
                    raise RuntimeError('Error on input argument list pair: ' + str(start_sample[index]) + '-' + str(end_sample[index]))

            # check range lengths
            if range_length.count(range_length[0]) != len(range_length):
                logging.error('The length of the ranges in the start_sample and end_sample arguments do not match. To retrieve multiple ranges/epochs, make sure that each of the start- and end-point combinations result in a range of the same length')
                raise RuntimeError('The length of the ranges in the start_sample and end_sample arguments do not match')

            # set output dimensions
            output_dimensions = [number_of_channels, range_length[0], len(range_length)]

        else:
            logging.error('Invalid input argument type(s). Both arguments should either be single values or lists/tuples of values')
            raise TypeError('Invalid input argument type(s)')

        # return whether single or multiple ranges and the output matrix dimensions and the ranges
        return single_range, output_dimensions, range_start, range_length



# define directory as package

#
import sys

# flatten access
from ieegprep.version import __version__
from ieegprep.fileio.IeegDataReader import VALID_FORMAT_EXTENSIONS

# ensure minimum version
if sys.version_info < (3, 8, 0):
    sys.exit("Python 3.8 or later is required.")


"""
Functions to read BrainVision files


=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

Adapted from Fieldtrip (by Robert Robert Oostenveld) while replicating some additional header logic from the MNE
package (Teon Brooks, Christian Brodbeck, Eric Larson, Jona Sassenhagen, Phillip Alday, Okba Bekhelifi, Stefan Appelhoff)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import sys
import os
import logging
import numpy as np
from configparser import ConfigParser
from .IeegDataReader import IeegDataReader
from ieegprep.utils.misc import allocate_array


class BrainVisionReader(IeegDataReader):

    bv_hdr = None                   # header
    bv_data = None                  # data (only used on preload)

    def __init__(self, data_path, data_preload=False):
        super().__init__(data_path, data_preload)
        self.data_format = 'bv'

        # load header
        try:
            self.bv_hdr = self.bv_read_header(self.data_path)
        except (FileNotFoundError, IOError):
            logging.error('Could not read BrainVision header')
            raise RuntimeError('Could not read BrainVision header')

        # retrieve the sample-rate, total number of samples and channel names
        self.sampling_rate = self.bv_hdr['sampling_rate']
        self.num_samples = self.bv_hdr['number_of_samples']
        self.channel_names = self.bv_hdr['channel_names']

        # (optionally) preload data
        if self.data_preload:
            try:
                _, self.bv_data = self.bv_read_data(filepath=self.data_path, hdr=self.bv_hdr)
            except (FileNotFoundError, IOError, TypeError, RuntimeError):
                logging.error('Could not read BrainVision data')
                raise RuntimeError('Could not read BrainVision data')


    def close(self):
        del self.bv_hdr
        if self.bv_data is not None:
            del self.bv_data


    def retrieve_channel_data(self, channel_name, ensure_own_data=True):
        """
        Retrieve the channel data (mef = numpy data-array, mne = numpy data-view)

        Args:
            channel_name:                   The name of the channel for which to retrieve the data
            ensure_own_data (bool):         Should ensure the return a numpy array has it's own data (is not a view)

        Returns:
            Numpy array with data

        Raises:
            LookupError:                    Raised when the channel name cannot be found
            RuntimeError:                   Raised when unable to retrieve channel data
        """

        # try to find the channel name
        try:
            channel_index = self.bv_hdr['channel_names'].index(channel_name)
        except ValueError:
            raise LookupError('Could not find channel')

        # determine whether the data is preloaded
        if self.bv_data is None:
            # not preloaded

            # load the channel data
            try:
                _, channel_data = self.bv_read_data(filepath=self.data_path,
                                                    hdr=self.bv_hdr,
                                                    channels=(channel_name,))
            except Exception:
                logging.error('Could not read BrainVision data')
                raise RuntimeError('Could not read BrainVision data')

            return channel_data.squeeze()

        else:
            # preloaded data

            # return the channel data
            if ensure_own_data:
                return self.bv_data[channel_index, :].copy()
            else:
                return self.bv_data[channel_index, :]


    def retrieve_sample_range_data(self, channels, sample_start, sample_end, ensure_own_data=True):
        """
        Retrieve a specific range of BrainVision data for the requested channels

        Args:
            channels (str, list or tuple):  The channel(s) for which to retrieve the data.
                                            If empty, all channels will be retrieved
            sample_start (int):             The start-point in time (in samples) to start reading from (0-based)
            sample_end (int):               The sample to end the reading at (0-based)
            ensure_own_data (bool):         Should ensure the return a numpy array has it's own data (is not a view)

        Returns:
            data                            A float data matrix containing the signal data (of the requested channels)
                                            formatted as <channels> x <samples/time>

        Raises:
            LookupError:                    Raised when a channel name cannot be found
            RuntimeError:                   Raised when unable to retrieve data
        """

        # check/prepare the channels argument
        if isinstance(channels, str):
            channels = [channels]
        if isinstance(channels, tuple):
            channels = list(channels)
        if channels is None or len(channels) == 0:
            channels = self.channel_names

        if self.bv_data is None:
            # data is not preloaded

            # load the data
            try:
                _, np_sample_data = self.bv_read_data(filepath=self.data_path, hdr=self.bv_hdr,
                                                      channels=channels,
                                                      start_sample=sample_start, end_sample=sample_end)
            except Exception:
                logging.error('Could not read BrainVision data')
                raise RuntimeError('Could not read BrainVision data')

            # return the data (from 2-d numpy array to list of numpy arrays)
            return list(np_sample_data)

        else:
            # data is preloaded

            # create a list with the numpy arrays
            sample_data = [None] * len(channels)

            # loop over the requested channels
            for counter in range(len(channels)):

                # find the channel index
                try:
                    channel_index = self.bv_hdr['channel_names'].index(channels[counter])
                except ValueError:
                    raise LookupError('Could not find channel')

                # pick the slice
                if ensure_own_data:
                    sample_data[counter] = self.bv_data[channel_index, sample_start:sample_end].copy()
                else:
                    sample_data[counter] = self.bv_data[channel_index, sample_start:sample_end]

        return sample_data


    #
    # BrainVision header and data reading functions
    #

    @staticmethod
    def bv_read_header(filepath) -> dict:
        """
        Read a BrainVision header (.vhdr) file

        Args:
            filepath (str):                 The path to the BrainVision header file

        Returns:
            A dictionary with header information

        Raises:
            FileNotFoundError:              Raised when unable to find the file
            IOError:                        Raised on any error that occurs while reading or parsing the header file

        Examples:
            hdr = bv_read_header('~/dataset.vhdr')
        """

        #
        # Read/check the file
        #

        # check file existence
        if not os.path.exists(filepath):
            logging.error('Header file \'' + filepath + '\' could not be found')
            raise FileNotFoundError('No such file or directory: \'' + filepath + '\'')

        # check extension
        if os.path.splitext(filepath)[1].lower() not in ('.vhdr',):
            logging.error('The header file \'' + filepath + '\' has an invalid extension (should be .vhdr)')
            raise IOError('Header file \'' + filepath + '\' has an invalid extension')

        # read
        with open(filepath, 'rb') as file:

            # retrieve and check the header
            header_line = file.readline().decode(encoding='ascii', errors='ignore').strip()
            if not header_line.startswith('BrainVision Data Exchange Header File Version ') and not \
                   header_line.startswith('Brain Vision Data Exchange Header File Version '):
                logging.error('Invalid or missing header in \'' + filepath + '\'')
                raise IOError('Invalid or missing header in .vhdr file')

            version = header_line[-3:]
            if not version == '1.0' and not version == '2.0':
                logging.error('Header file indicates an unknown BrainVision format version \'' + filepath + '\'')
                raise IOError('Unknown BrainVision format version')

            # parse the rest of the INI file
            try:
                config = ConfigParser()
                config.read_string(file.read().decode(encoding='UTF-8', errors='ignore'))
            except:
                logging.error('Error while reading and parsing the BrainVision header file \'' + filepath + '\'')
                raise IOError('Unable to read/parse .vhdr file')


        #
        # Obtain information from the header
        #

        def getField(parser, section, field, type, optional=False):
            try:
                if type == 'str':
                    return parser.get(section, field)
                elif type == 'int':
                    return parser.getint(section, field)
                elif type == 'float':
                    return parser.getfloat(section, field)
            except:
                if optional:
                    return None
                else:
                    logging.error('Could not read/parse mandatory option \'' + field + '\' (section \'' + section + '\') from the BrainVision header file \'' + filepath + '\'')
                    raise IOError('Unable to read/parse mandatory option from .vhdr file')

        hdr = dict()
        try:
            hdr['data_file']            = getField(config, 'Common Infos', 'DataFile', 'str')
            hdr['number_of_channels']   = getField(config, 'Common Infos', 'NumberOfChannels', 'int')
            hdr['Sampling_interval']    = getField(config, 'Common Infos', 'SamplingInterval', 'float')
            hdr['sampling_rate']        = 1e6 / hdr['Sampling_interval']
            hdr['data_points']          = getField(config, 'Common Infos', 'DataPoints', 'int', True)

            # format
            hdr['data_format']          = getField(config, 'Common Infos', 'DataFormat', 'str')
            if hdr['data_format'] == 'BINARY':
                hdr['data_binary_format']          = getField(config, 'Binary Infos', 'BinaryFormat', 'str')
                if not hdr['data_binary_format'] in ('INT_16', 'INT_32', 'IEEE_FLOAT_32'):
                    logging.error('Unknown data binary format \'' + hdr['data_binary_format'] + '\' in BrainVision header file \'' + filepath + '\'')
                    raise IOError('Unknown data binary format in .vhdr file')

            elif hdr['data_format'] == 'ASCII':

                skip_lines                          = getField(config, 'ASCII Infos', 'SkipLines', 'int', True)
                hdr['data_ascii_skip_lines']        = skip_lines if skip_lines is not None else 0
                skip_columns                        = getField(config, 'ASCII Infos', 'SkipColumns', 'int', True)
                hdr['data_ascii_skip_columns']      = skip_columns if skip_columns is not None else 0
                hdr['data_ascii_decimal_symbol']    = getField(config, 'ASCII Infos', 'DecimalSymbol', 'str', True)

            else:
                logging.error('Unknown data format \'' + hdr['data_format'] + '\' in BrainVision header file \'' + filepath + '\'')
                raise IOError('Unknown data format in .vhdr file')

            # orientation
            hdr['data_orientation']     = getField(config, 'Common Infos', 'DataOrientation', 'str')
            if not hdr['data_orientation'] in ('MULTIPLEXED', 'VECTORIZED'):
                logging.error('Unknown data orientation \'' + hdr['data_orientation'] + '\' in BrainVision header file \'' + filepath + '\'')
                raise IOError('Unknown data orientation in .vhdr file')


            #
            # channel information
            #

            hdr['channel_names']             = []
            hdr['channel_other_info']        = []
            if hdr['number_of_channels'] != 0:
                for i in range(1, hdr['number_of_channels'] + 1):

                    # retrieve the channel info
                    chan_str  = ('Ch' + str(i))
                    chan_info = getField(config, 'Channel Infos', chan_str, 'str').split(',')

                    # Note: It is not clear whether the resolution and unit are related in the Brainvision format. It
                    #       could be that they need to match (where the resolution is the conversion factor to get to
                    #       the unit). However, the 'BrainVision Core Data Format' documentation also seems to mention
                    #       an example that speaks of 'Resolution / Unit' with reported values of '0.5 μV'. These imply
                    #       that resolution and unit are independent values.

                    # if no resolution is set, then set resolution to 1 (as specified in 'BrainVision Core Data Format' documentation)
                    chan_res = float(chan_info[2]) if chan_info[2] == '' else 1

                    # if no unit is available, then set unit to µV (as specified in 'BrainVision Core Data Format' documentation)
                    chan_unit = str(chan_info[3]) if len(chan_info) > 3 and not chan_info[3] == '' else 'µV'

                    # determine the unit for specific unit codes (otherwise ensure multiplication by 1)
                    # Note: setting this unit is used to later allow us to convert to the requested output (uV, mV or V)
                    if chan_unit.lower() in ('\u03BCv', '\u00B5v', '\x83\xCAv', 'uv'):
                        units_to_v_gain = 1e-6
                    elif chan_unit.lower() == 'mv':
                        units_to_v_gain = 1e-3
                    elif chan_unit.lower() == 'v':
                        units_to_v_gain = 1
                    else:
                        #logging.warning('The unit on channel ' + hdr['channel_names'][count] + ' is not unknown, assume multiplication by 1')
                        units_to_v_gain = 1

                    # store (and remove some troublesome characters)
                    hdr['channel_names'].append(chan_info[0].replace(r'\1', ','))
                    hdr['channel_other_info'].append({"ref": chan_info[1],
                                                      "resolution": chan_res,
                                                      "unit": chan_unit.replace('\xc2', ''),
                                                      "units_to_V_gain": units_to_v_gain})

        except:
            logging.error('Unable to obtain the required fields from the BrainVision header file \'' + filepath + '\'')
            raise IOError('Unable to obtain the required field from the .vhdr file')


        #
        # determine the number of samples
        #

        # determine data filepath
        base_dir = os.path.dirname(filepath)
        data_file = os.path.splitext(os.path.basename(filepath))[0] + '.eeg'
        if not data_file == hdr['data_file']:
            logging.warning('The name of the data-file, based on the header\'s filename (\'' + data_file + '\') and DataFile field (\'' + hdr['data_file'] + '\'), are not in agreement.\nUsing name based on header\'s filename')
        data_file = os.path.join(base_dir, data_file)

        # infer number of samples
        if hdr['data_format'] == 'BINARY':

            # determine the size of the data file
            try:
                with open(data_file, 'rb') as file:
                    num_bytes = file.seek(0, os.SEEK_END)
            except:
                logging.error('Error while getting the file size of the BrainVision data file \'' + data_file + '\', unable to determine number of samples')
                raise IOError('Unable to the file size of the .eeg file, unable to determine number of samples')

            # calculate the number of samples
            if hdr['data_binary_format'] == 'INT_16':
                hdr['number_of_samples'] = int(num_bytes / (hdr['number_of_channels'] * 2))
            elif hdr['data_binary_format'] == 'INT_32' or hdr['data_binary_format'] == 'IEEE_FLOAT_32':
                hdr['number_of_samples'] = int(num_bytes / (hdr['number_of_channels'] * 4))

        elif hdr['data_format'] == 'ASCII':

            if hdr['data_points'] is not None:
                hdr['number_of_samples'] = hdr['data_points']
            else:

                if hdr['data_orientation'] == 'VECTORIZED':
                    logging.warning('No DataPoints field in the header file, inferring number of samples from the (ASCII) data-file assuming VECTORIZED data orientation')

                    # determine the number of values in the first line
                    try:
                        with open(data_file, 'r') as file:
                            first_line = file.readline().decode(encoding='ascii', errors='ignore')
                    except:
                        logging.error('Error while reading BrainVision data file \'' + data_file + '\', unable to determine number of samples')
                        raise IOError('Error while reading the .eeg file, unable to determine number of samples')

                    # infer the number of samples
                    first_line_values = first_line.replace('  ', ' ').split(' ')
                    if len(first_line_values) < 2:
                        logging.error('Error while inferring the number of samples, the first line in the data file has no values')
                        raise IOError('Error while inferring the number of samples, the first line in the data file has no values')
                    hdr['number_of_samples'] = int(len(first_line_values) - 1)

                else:
                    # ASCII MULTIPLEXED
                    logging.warning('No DataPoints field in the header file and cannot infer number of samples from the (ASCII) data-files with a MULTIPLEXED orientation')

                    # count the lines
                    try:
                        num_lines = 0
                        for _ in open(data_file, 'r'):
                            num_lines += 1
                    except:
                        logging.error('Error while reading BrainVision data file \'' + data_file + '\', unable to determine number of lines/samples')
                        raise IOError('Error while reading the .eeg file, unable to determine number of lines/samples')

                    # set the number of samples as the number of lines minus the number of lines to skip, minus 1
                    hdr['number_of_samples'] = int(num_lines - hdr['data_ascii_skip_lines'] - 1)

        # TODO: more information like annotations, coordinates etc

        # return the header information
        return hdr


    @staticmethod
    def bv_read_data(filepath, hdr=None, channels=None, start_sample=0, end_sample=-1, unit='uV', use_memmap=True, chunked_read=False) -> tuple:
        """
        Read data from a BrainVision (.eeg) file

        Args:
            filepath (str):                 The path to the BrainVision data file
            hdr (dict):                     Optionally pass the header dictionary as provided by 'bv_read_header'. If this
                                            argument is not set, the header file will be parsed before retrieving the data.
                                            If set, this function will use the header that is provided (to prevent multiple
                                            header reads on multiple calls to this function)
            channels (str/list/tuple):      The names of the channels to return the signal data from. The order of channels
                                            in this input argument will determine the order of channels in the output matrix.
                                            If set to None (default), all channels will be read and ordered according to the
                                            header file (hdr['channel_names'])
            start_sample (int or list):     The start-point in time (in samples) to start reading from (0-based). This argument
                                            can either be a single value to indicate the start value of a single range that
                                            should be retrieved over all the requested channels; or this argument can be a
                                            list of start values to retrieve multiple ranges (start- and end-points) for all
                                            the requested channels. When retrieving multiple ranges, make sure the number of
                                            values in the start_sample and the end_sample arguments match, and the size of the
                                            ranges (in samples) are equal.
            end_sample (int or list):       The sample to end the reading (0-based). This argument can either be a single
                                            value to indicate the end-sample of a single range that should be retrieved
                                            over all the requested channels; or this argument can be a list of end values
                                            to retrieve multiple ranges (start- and end-points) for all the requested
                                            channels. When retrieving multiple ranges, make sure the number of values in
                                            the start_sample and the end_sample arguments match, and the size of the ranges
                                            (in samples) are equal. Also note that because the range is 0-based, data are
                                            loaded "up-till" the range end-index, so the result does not include the value
                                            at the end-index (e.g. a requested sample range of 0-3 will return the first 3 values,
                                            being the values at [0], [1], [2]). A value of -1 (default) or -1 value in the list
                                            of values represents the latest sample of the time-series (as indicated by the header)
            unit (str):                     The unit in which the data should be return ('uV' returns microVolts, 'mV' returns
                                            microVolts and 'V' returns Volts)
            use_memmap (bool):              Whether to use numpy's memmap (which wraps around mmap) while reading the data.
                                            If true, the data file is first loaded/cached into virtual memory (pagefile) to
                                            speed up (repetitive) reading. If false, the standard system read operations
                                            will be used (slightly slower for repetitive reading from the file but does not
                                            explicitly require virtual memory).
            chunked_read (bool):            Whether to read or transfer the data in chunks. If true, then reads from
                                            disk (stdIO) or transfers from virtual memory (memmap) are split up in xxMB
                                            chunks. If false, then often all data is read or transferred in one big block.

        Returns:
            hdr                             A dictionary with header information
            data                            A float data matrix containing the signal data (of the requested channels). When
                                            a single time-series is requested (one value as start_sample and one value as
                                            end_sample argument) then the matrix will be formatted as <channels> x <samples/time>,
                                            with the first dimension (rows) representing the channels (ordered based on
                                            the 'channels' input argument) and the second dimension (columns) the samples/time.
                                            If multiple ranges are requested (lists of values for the start_sample and end_sample
                                            arguments) then the return format will be <channels> x <samples/time> x <ranges>, so
                                            that the third dimension represents the requested ranges/epochs.

        Raises:
            FileNotFoundError:              Raised when unable to find the file
            IOError:                        Raised on any error that occurs while reading or parsing the data file
            TypeError:                      Raised when input argument types are wrong
            RuntimeError:                   Raised upon a processing error

        Examples:
            hdr, data = bv_read_data('~/dataset.eeg')
            hdr, data = bv_read_data('~/dataset.eeg', channels='CH01')
            hdr, data = bv_read_data('~/dataset.eeg', channels=('CH01', 'CH07'))
            hdr, data = bv_read_data('~/dataset.eeg', start_sample=1000, end_sample=2000)
            hdr, data = bv_read_data('~/dataset.eeg', start_sample=(1000, 6000), end_sample=(2000, 7000))
        """

        # check unit argument
        if unit.lower() not in ('uv', 'mv', 'v'):
            logging.error('Invalid unit ' + unit + ' to retrieve the data in. Only options are uV, mV or V')
            raise RuntimeError('Invalid unit')

        # check file existence
        if not os.path.exists(filepath):
            logging.error('Data file \'' + filepath + '\' could not be found')
            raise FileNotFoundError('No such file or directory: \'' + filepath + '\'')

        # check whether to load the header
        if hdr is None:

            # read the header information
            header_file = os.path.splitext(filepath)[0] + '.vhdr'
            try:
                hdr = BrainVisionReader.bv_read_header(header_file)
            except (FileNotFoundError, IOError):
                logging.error('Error while reading/parsing header file \'' + header_file + '\', this information is required to read the data file')
                raise IOError('Error while reading/parsing header file, this info required to read data file')

        else:
            # TODO: check fields in the header file that is passed?
            pass

        # check whether there are samples
        if hdr['number_of_samples'] == 0:
            logging.error('no samples in the data file according to the header')
            raise RuntimeError('no samples in the data')


        #
        # check/prepare channel input(s)
        #

        if isinstance(channels, str):
            channels = (channels,)

        # retrieve channel indices
        if channels is None or len(channels) == 0:
            channels = hdr['channel_names']
            channel_indices = list(range(0, len(channels)))
        else:
            channel_indices = []
            for channel in channels:
                try:
                    index = hdr['channel_names'].index(channel)
                    channel_indices.append(index)
                except ValueError:
                    logging.error('requested channel \'' + channel + '\' was not found')
                    raise IOError('Requested channel was not found')

        # list the gains of the channels in a list (for speed), consider the resolution and unit, also
        # determine whether perhaps all channels have a gain/multiplication of 1
        all_gains_one = True
        channel_gains = []
        for channel_idx in channel_indices:

            # retrieve the gain that needs to be applied to get the data in the requested unit
            unit_gain = hdr['channel_other_info'][channel_idx]['units_to_V_gain']
            if unit.lower() == 'uv':
                unit_gain *= 1e+6
            elif unit.lower() == 'mv':
                unit_gain *= 1e+3

            # calculate and store the total gain (based on resolution and unit)
            gain = hdr['channel_other_info'][channel_idx]['resolution'] * unit_gain
            channel_gains.append(gain)

            if gain != 1:
                all_gains_one = False

        channel_gains = tuple(channel_gains)

        #
        # check/prepare sample range input(s)
        #

        single_range, output_dimensions, ranges_start, ranges_length = IeegDataReader.check_sample_arguments(start_sample, end_sample,
                                                                                                             hdr['number_of_samples'],
                                                                                                             len(channel_indices))

        #
        # Read/check the data file
        #
        data_file = os.path.splitext(filepath)[0] + '.eeg'

        # check file existence
        if not os.path.exists(data_file):
            logging.error('Data file \'' + data_file + '\' could not be found')
            raise FileNotFoundError('No such file or directory: \'' + data_file + '\'')

        # determine the sample size in bytes
        if hdr['data_format'] == 'BINARY':
            if hdr['data_binary_format'] == 'INT_16':
                sample_size = 2
                sample_type = '<i2'
            elif hdr['data_binary_format'] == 'INT_32':
                sample_size = 4
                sample_type = '<i4'
            elif hdr['data_binary_format'] == 'IEEE_FLOAT_32':
                sample_size = 4
                sample_type = '<f4'

        # check if the system is 32-bit and memmap is enabled
        # (mapping on 32-bit systems is limited to 2GB files, so disable just in case)
        if use_memmap and not (sys.maxsize > 2 ** 32):
            logging.warning('This is a 32-bit system. Data will be read with standard IO operations (rather than memmap)')
            use_memmap = False

        #
        if hdr['data_format'] == 'BINARY' and hdr['data_binary_format'] in ('INT_16', 'INT_32', 'IEEE_FLOAT_32'):

            # init data matrix (with a specific output data-type)
            if all_gains_one:
                # all channel resolutions are 1, so no multiplication with a fractional number here.
                # since manipulations of the output data are expected to involve fractional precision, the output matrix
                # will be of the float datatype. Whether this is float32 or float64 depends on the data-type in the data file

                if hdr['data_binary_format'] == 'INT_16':
                    # int16 will be cast to float32 (using less memory than a float64). Float32 has 24-bit significand
                    # precision and can therefore easily hold the exact values in the range of int16
                    data = np.empty(output_dimensions, dtype=np.float32)
                    #allocate_array(output_dimensions, fill_value=np.nan, dtype=np.float32)

                elif hdr['data_binary_format'] == 'INT_32':
                    # int32 will be cast to float64. Float64 has 53-bit significand precision and can therefore hold the
                    # exact values in the range of int32 (float32 would not be able to)
                    data = np.empty(output_dimensions, dtype=np.float64)
                    #allocate_array(output_dimensions, fill_value=np.nan, dtype=np.float64)

                elif hdr['data_binary_format'] == 'IEEE_FLOAT_32':
                    # Keep in original float32 data type. Casting to float64 is possible but at this point would only add
                    # precision that is not really in the data while doubling the memory footprint.
                    data = np.empty(output_dimensions, dtype=np.float32)
                    #allocate_array(output_dimensions, fill_value=np.nan, dtype=np.float32)

            else:
                # at least one resolution is not 1, so a multiplication with - possibly - a fractional number might occur.
                # To facilitate as much precision as possible we will work with float64 numbers (at the expense of more memory)
                data = np.empty(output_dimensions, dtype=np.float64)
                #allocate_array(output_dimensions, fill_value=np.nan, dtype=np.float64)


            # local variable for number of channels
            num_channels_in_file = len(hdr['channel_names'])

            #
            if hdr['data_orientation'] == 'MULTIPLEXED':

                # Conclusions
                # - when reading samples, only part is cached
                # - when reading all or a single channel, the entire file is cached
                #
                # - after caching:
                #   - chunked reading always faster than not chunked reading, and smaller chunk sizes are faster
                #   - memmap is faster than fromfile

                # prevent per sample reading (due to reading non-chunked, standard IO and picking channels) because
                # it is notoriously slow, switch to chunked instead
                if not chunked_read and not use_memmap and not (channel_indices == list(range(0, num_channels_in_file))):
                    logging.warning('The current read settings (non-chunked, standard IO and picking channels will result in per sample reading, this is highly discouraged and exists for testing purposes only. Switching to chunked reading instead')
                    chunked_read = True

                #
                if chunked_read:
                    # chunked reading
                    #print('multiplexed - chunked - memmap = ' + str(use_memmap))

                    if use_memmap:
                        amem = np.memmap(data_file, dtype=sample_type, mode='r', offset=0, order='C')
                    else:
                        try:
                            fid_data = open(data_file, "rb")
                        except IOError:
                            logging.error('Error while opening data file \'' + data_file + '\'')
                            raise IOError('Error while opening data file')

                    # determine the size (in samples) chucks assuming ~10mb per read
                    # rounded down to the number of channels
                    chunk_size = ((int(10e6) // sample_size) // num_channels_in_file) * num_channels_in_file
                    assert (chunk_size != 0)

                    def read_binary_chunked(rb_chunk_size, rb_range_index, rb_start, rb_length):

                        # determine where to start reading from
                        if not use_memmap:
                            fid_data.seek(rb_start * num_channels_in_file * sample_size, os.SEEK_SET)

                        #
                        chunk_start = rb_start * num_channels_in_file
                        data_sample_pos = 0
                        while data_sample_pos < rb_length:

                            # determine length (in samples) of the current chunk
                            current_chunk_length = min((rb_length - data_sample_pos) * num_channels_in_file, rb_chunk_size)

                            if use_memmap:
                                chunk_data = amem[chunk_start:chunk_start + current_chunk_length].reshape(num_channels_in_file, -1, order='F')
                                chunk_start += current_chunk_length
                            else:
                                chunk_data = np.fromfile(fid_data, dtype=sample_type, count=current_chunk_length).reshape(num_channels_in_file, -1, order='F')
                            num_samples = chunk_data.shape[1]

                            # pick the channels and (optionally) apply the resolution
                            # Note 1. will automatically cast data-types (int->float) later if needed
                            # Note 2. deliberate more ifs, because function calls are expensive in python (performance wise) and these are iterated through a lot
                            for channel_count, channel_index in enumerate(channel_indices):
                                if channel_gains[channel_count] == 1:
                                    if rb_range_index is None:
                                        data[channel_count, data_sample_pos:data_sample_pos + num_samples] = chunk_data[channel_index, :]
                                    else:
                                        data[channel_count, data_sample_pos:data_sample_pos + num_samples, rb_range_index] = chunk_data[channel_index, :]
                                else:
                                    if rb_range_index is None:
                                        data[channel_count, data_sample_pos:data_sample_pos + num_samples] = chunk_data[channel_index, :].astype(np.float64) * channel_gains[channel_count]
                                    else:
                                        data[channel_count, data_sample_pos:data_sample_pos + num_samples, rb_range_index] = chunk_data[channel_index, :].astype(np.float64) * channel_gains[channel_count]

                            # move to the next chunk
                            data_sample_pos += num_samples

                    # read range(s)
                    if single_range:
                        read_binary_chunked(chunk_size, None, ranges_start, ranges_length)
                    else:
                        for range_index in range(len(ranges_start)):
                            read_binary_chunked(chunk_size, range_index, ranges_start[range_index], ranges_length[range_index])

                    # close file if needed
                    if not use_memmap and fid_data is not None:
                        fid_data.close()

                else:
                    # non-chunked reading

                    # check if all channels & in the same order as in the file (which means that we can just retrieve a blocks of samples and transpose the matrix)
                    if channel_indices == list(range(0, num_channels_in_file)):
                        #print('multiplexed - non-chunked - memmap = ' + str(use_memmap) + ' - allchan & sameorder')

                        # open
                        if use_memmap:
                            amem = np.memmap(data_file, dtype=sample_type, mode='r', offset=0, order='C')
                        else:
                            try:
                                fid_data = open(data_file, "rb")
                            except IOError:
                                logging.error('Error while opening data file \'' + data_file + '\'')
                                raise IOError('Error while opening data file')

                        #
                        def read_binary_multiplexed(rb_start, rb_length):
                            # will automatically cast data-types (int->float) later if needed
                            chan_start = rb_start * num_channels_in_file
                            if use_memmap:
                                return amem[chan_start:chan_start + rb_length * num_channels_in_file].reshape(num_channels_in_file, rb_length, order='F')
                            else:
                                fid_data.seek(chan_start * sample_size, os.SEEK_SET)
                                return np.fromfile(fid_data, dtype=sample_type, count=rb_length * num_channels_in_file).reshape(num_channels_in_file, -1, order='F')

                        # loop over the ranges
                        if single_range:
                            data[:, :] = read_binary_multiplexed(ranges_start, ranges_length)
                        else:
                            for range_index in range(len(ranges_start)):
                                data[:, :, range_index] = read_binary_multiplexed(ranges_start[range_index], ranges_length[range_index])

                        # apply the resolution per channel
                        for channel_count, channel_index in enumerate(channel_indices):
                            if not channel_gains[channel_count] == 1:
                                if single_range:
                                    data[channel_count, :] *= channel_gains[channel_count]
                                else:
                                    data[channel_count, :, :] *= channel_gains[channel_count]

                    else:
                        # not all channels or not in the same order as file
                        #print('multiplexed - non-chunked - memmap = ' + str(use_memmap) + ' - not allchan or not sameorder')

                        if use_memmap:

                            # open memory map and define the read function
                            amem = np.memmap(data_file, dtype=sample_type, mode='r', offset=0, shape=(num_channels_in_file, hdr['number_of_samples']), order='F')

                            def read_binary_multiplexed(rb_chan_index, rb_start, rb_length):
                                if channel_gains[channel_count] == 1:
                                    return amem[channel_index, rb_start:rb_start + rb_length]  # will automatically cast data-types (int->float) later if needed
                                else:
                                    return amem[channel_index, rb_start:rb_start + rb_length].astype(np.float64) * channel_gains[channel_count]

                            # loop over the channels/ranges
                            for channel_count, channel_index in enumerate(channel_indices):
                                if single_range:
                                    data[channel_count, :] = read_binary_multiplexed(channel_index, ranges_start, ranges_length)
                                else:
                                    for range_index in range(len(ranges_start)):
                                        data[channel_count, :, range_index] = read_binary_multiplexed(channel_index, ranges_start[range_index], ranges_length[range_index])


                        else:
                            # no other choice but to loop over samples and over channels (since channels are not always consecutive)

                            logging.warning('Should not be used, rather read in chunks')

                            try:
                                fid_data = open(data_file, "rb")
                            except IOError:
                                logging.error('Error while opening data file \'' + data_file + '\'')
                                raise IOError('Error while opening data file')

                            #
                            def read_binary_multiplexed(rb_range_index, rb_start, rb_length):
                                data_sample_pos = 0
                                if rb_range_index is None:

                                    for sample_index in range(rb_start, rb_start + rb_length):
                                        for channel_count, channel_index in enumerate(channel_indices):
                                            fid_data.seek((sample_index * num_channels_in_file + channel_index) * sample_size, os.SEEK_SET)
                                            if channel_gains[channel_count] == 1:
                                                data[channel_count, data_sample_pos] = np.fromfile(fid_data, dtype=sample_type, count=1)
                                            else:
                                                data[channel_count, data_sample_pos] = np.fromfile(fid_data, dtype=sample_type, count=1).astype(np.float64) * channel_gains[channel_count]
                                        data_sample_pos += 1

                                else:

                                    for sample_index in range(rb_start, rb_start + rb_length):
                                        for channel_count, channel_index in enumerate(channel_indices):
                                            fid_data.seek((sample_index * num_channels_in_file + channel_index) * sample_size, os.SEEK_SET)
                                            if channel_gains[channel_count] == 1:
                                                data[channel_count, data_sample_pos, range_index] = np.fromfile(fid_data, dtype=sample_type, count=1)
                                            else:
                                                data[channel_count, data_sample_pos, range_index] = np.fromfile(fid_data, dtype=sample_type, count=1).astype(np.float64) * channel_gains[channel_count]
                                        data_sample_pos += 1


                            # loop over the ranges
                            if single_range:
                                read_binary_multiplexed(None, ranges_start, ranges_length)
                            else:
                                for range_index in range(len(ranges_start)):
                                    read_binary_multiplexed(range_index, ranges_start[range_index], ranges_length[range_index])


            elif hdr['data_orientation'] == 'VECTORIZED':

                # Conclusions
                # - when reading samples or a single channel, only part is cached
                # - when reading all, the entire file is cached
                #
                # - after caching:
                #   - on memmap chunked read is faster, on fromfile non-chunked read is faster
                #   - in general memmap is faster than fromfile

                if use_memmap:
                    amem = np.memmap(data_file, dtype=sample_type, mode='r', offset=0, order='C')
                else:
                    try:
                        fid_data = open(data_file, "rb")
                    except IOError:
                        logging.error('Error while opening data file \'' + data_file + '\'')
                        raise IOError('Error while opening data file')

                if chunked_read:
                    # chunked reading
                    #print('vectorized - chunked - memmap = ' + str(use_memmap))

                    # determine the size (in samples) chucks assuming ~10mb per read
                    # rounded down to the number of channels (also here because than we ensure that we assign in chunks as a multiple of the channel dimension)
                    chunk_size = ((int(10e6) // sample_size))
                    assert(chunk_size != 0)

                    # read function
                    def read_binary_chunked(rb_chunk_size, rb_range_index, rb_start, rb_length):

                        # Note: reading each channel in chunks (since there is no guarantee that the requested channels are consecutive)
                        #       the chunk can be shortened depending on the number of samples in the channel or the requested range

                        # channels loop
                        for channel_count, channel_index in enumerate(channel_indices):

                            # (determine where to) start reading from
                            if use_memmap:
                                channel_sample_start = channel_index * hdr['number_of_samples'] + rb_start
                            else:
                                fid_data.seek((channel_index * hdr['number_of_samples'] + rb_start) * sample_size, os.SEEK_SET)

                            # chunk loop
                            data_sample_pos = 0
                            while data_sample_pos < rb_length:

                                # determine length (in samples) of the current chunk
                                current_chunk_length = min((rb_length - data_sample_pos), rb_chunk_size)

                                # read the chunk
                                if use_memmap:
                                    chunk_start = channel_sample_start + data_sample_pos

                                # transfer chunk data
                                # Note 1. will automatically cast data-types (int->float) later if needed
                                # Note 2. deliberate more ifs, because function calls are expensive in python (performance wise) and these are iterated through a lot
                                if channel_gains[channel_count] == 1:
                                    if rb_range_index is None:
                                        if use_memmap:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length] = amem[chunk_start:chunk_start + current_chunk_length]
                                        else:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length] = np.fromfile(fid_data, dtype=sample_type, count=current_chunk_length)
                                    else:
                                        if use_memmap:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length, rb_range_index] = amem[chunk_start:chunk_start + current_chunk_length]
                                        else:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length, rb_range_index] = np.fromfile(fid_data, dtype=sample_type, count=current_chunk_length)
                                else:
                                    if rb_range_index is None:
                                        if use_memmap:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length] = amem[chunk_start:chunk_start + current_chunk_length].astype(np.float64) * channel_gains[channel_count]
                                        else:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length] = np.fromfile(fid_data, dtype=sample_type, count=current_chunk_length).astype(np.float64) * channel_gains[channel_count]
                                    else:
                                        if use_memmap:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length, rb_range_index] = amem[chunk_start:chunk_start + current_chunk_length].astype(np.float64) * channel_gains[channel_count]
                                        else:
                                            data[channel_count, data_sample_pos:data_sample_pos + current_chunk_length, rb_range_index] = np.fromfile(fid_data, dtype=sample_type, count=current_chunk_length).astype(np.float64) * channel_gains[channel_count]

                                # move to the next chunk
                                data_sample_pos += current_chunk_length

                    # read range(s)
                    if single_range:
                        read_binary_chunked(chunk_size, None, ranges_start, ranges_length)
                    else:
                        for range_index in range(len(ranges_start)):
                            read_binary_chunked(chunk_size, range_index, ranges_start[range_index], ranges_length[range_index])

                else:
                    # non-chunked reading
                    #print('vectorized - non-chunked - memmap = ' + str(use_memmap))

                    # read function
                    # TODO: short function, move into loop below? Performance test to decide
                    if use_memmap:
                        def read_binary_vectorized(rb_chan_count, rb_start, rb_length):
                            if channel_gains[rb_chan_count] == 1:
                                return amem[rb_start:rb_start + rb_length]
                            else:
                                return amem[rb_start:rb_start + rb_length].astype(np.float64) * channel_gains[channel_count]
                    else:
                        def read_binary_vectorized(rb_chan_count, rb_start, rb_length):
                            fid_data.seek(rb_start * sample_size, os.SEEK_SET)
                            if channel_gains[rb_chan_count] == 1:
                                return np.fromfile(fid_data, dtype=sample_type, count=rb_length)
                            else:
                                return np.fromfile(fid_data, dtype=sample_type, count=rb_length).astype(np.float64) * channel_gains[channel_count]

                    # loop over the channels/ranges
                    for channel_count, channel_index in enumerate(channel_indices):
                        channel_sample_start = channel_index * hdr['number_of_samples']

                        # will automatically cast data-types (int->float) if needed
                        if single_range:
                            data[channel_count, :] = read_binary_vectorized(channel_count, channel_sample_start + ranges_start, ranges_length)
                        else:
                            for range_index in range(len(ranges_start)):
                                data[channel_count, :, range_index] = read_binary_vectorized(channel_count, channel_sample_start + ranges_start[range_index], ranges_length[range_index])

            # close file if needed
            if not use_memmap and fid_data is not None:
                fid_data.close()


        elif hdr['data_format'] == 'ASCII' and hdr['data_orientation'] == 'MULTIPLEXED':
            logging.error('Reading ASCII format with MULTIPLEXED orientation is not supported, test data is required')
            raise IOError('Reading ASCII format with MULTIPLEXED orientation is not supported, test data is required')

        elif hdr['data_format'] == 'ASCII' and hdr['data_orientation'] == 'VECTORIZED':
            logging.error('Reading ASCII format with VECTORIZED orientation is not supported, test data is required')
            raise IOError('Reading ASCII format with VECTORIZED orientation is not supported, test data is required')

        else:
            logging.error('Unknown data format')
            raise IOError('Unknown data format')

        # return the header information and the data
        return hdr, data



"""
Miscellaneous functions and classes

=====================================================
Copyright 2023, Max van den Boom (Multimodal Neuroimaging Lab, Mayo Clinic, Rochester MN)

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import logging
import numpy as np
import time
import subprocess
from ieegprep.utils.console import ConsoleColors


def allocate_array(dimensions, fill_value=np.nan, dtype=np.float64):
    """
    Create and immediately allocate the memory for an x-dimensional array

    Before allocating the memory, this function checks if is enough memory is available (this is needed since when a
    numpy array is allocated and there is not enough memory, sometimes python crashes without the chance to catch an error).

    Args:
        dimensions (int or tuple):
        fill_value (any numeric):
        dtype (str):

    Returns:
        data (ndarray):             An initialized x-dimensional array, or None if insufficient memory available

    """
    # initialize a data buffer
    mem = None
    try:

        # create a ndarray object (no memory is allocated here)
        data = np.empty(dimensions, dtype=dtype)
        data_bytes_needed = data.nbytes

        # import here to decrease the package dependencies for this module
        from psutil import virtual_memory

        # check if there is enough memory available
        mem = virtual_memory()
        if mem.available <= data_bytes_needed:
            raise MemoryError()

        # allocate the memory
        data.fill(fill_value)

        #
        return data

    except MemoryError:
        if mem is None:
            logging.error('Not enough memory available to create array.\n(for docker users: extend the memory resources available to the docker service)')
        else:
            logging.error('Not enough memory available to create array.\nAt least ' + str(int((mem.used + data_bytes_needed) / (1024.0 ** 2))) + ' MB total is needed, most likely more.\n(for docker users: extend the memory resources available to the docker service)')
        raise MemoryError('Not enough memory available to create array.')


def is_number(value):
    try:
        float(value)
        return True
    except:
        return False


def run_cmd(command, env={}):
    merged_env = os.environ
    merged_env.update(env)
    merged_env.pop('DEBUG', None)
    process = subprocess.run(command,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             shell=True,
                             universal_newlines=True,
                             env=merged_env,
                             encoding='utf-8')
    return process


def time_func(fun, pre_fun = None, loop = 5, *args, **kwargs):
    """
    Time and retrieve statistics on the performance of a given function

    Args:
        fun (function):                 The function to time the performance on
        pre_fun (function):             Function that is called every time prior to calling the function being
                                        timed, but not counted as part of the performance time
        loop (int):                     Number of times the function is called
        args (*args):
        kwargs (**kwargs):

    Returns:
        mean:                       The function's average execution time (in ms)
        std:                        The standard deviation in the function's execution time
        range (tuple):              The range of the function's execution time. The first value is minimum, second is the maximum
        times (nparray):            All result values

    """
    if loop < 1:
        loop = 1
    times = np.zeros(loop)

    # repeatedly call function
    for iLoop in range(0, loop):

        # execute the pre-function if there is one
        if pre_fun :
            pre_fun()

        # execute function and measure
        time1 = time.time()
        fun(*args, **kwargs)
        time2 = time.time()

        # store
        times[iLoop] = (time2 - time1) * 1000.0

    # return statistics
    return times.mean(), times.std(), (times.min(), times.max()), times


def clear_virtual_cache():
    """
    Try to clear the virtual memory (pagefile)
    """

    from sys import platform
    if platform == "win32":
        # Note: Running the RAMMAP executable will usually popup a User Access Control (UAC), warning about changes
        #       being made to the computer. This can be prevented by either temporarily setting the UAC
        #       notification setting to off (don't forget to turn it back on after) or to disable UAC for
        #       only the RAMMap64 executable

        # check whether the executable exits
        if not os.path.exists('RAMMap64.exe'):
            ConsoleColors.print_error('Error: could not find RAMMAP64.exe to clear virtual memory.\nDownload RAMMAP tools from Microsoft Sysinternals, and make sure the RAMMAP64.exe file is into the script directory (or can be found through the environment variable $PATH)')
            exit(1)

        # call clear executable
        os.system('RAMMap64.exe -Et')
        print('Cleared virtual memory')

    elif platform in ("linux", "linux2"):

        os.sync()
        with open('/proc/sys/vm/drop_caches', 'w') as f:
            f.write("1\n")
        print('Cleared virtual memory')

    elif platform == "darwin":
        # Note: Executing the command 'purge' to clear the virtual memory should prompt for a password, but
        #       it is better to edit the 'sudoers' file, so that the 'purge' command can be executed without the
        #       admin password. To achieve this:
        #        1. In a terminal type 'sudo visudo' to edit the 'sudoers' file
        #        2. At the end of that file add a line:
        #           <mac username> ALL = NOPASSWD: /usr/sbin/purge
        #           e.g. john ALL = NOPASSWD: /usr/sbin/purge
        #        3. save and close the file

        # create bash file to GUI prompt for a password
        with open('./mac_prompt_pw.sh', 'w') as f:
            f.write('#!/bin/bash\n')
            f.write('pw="$(osascript -e \'Tell application "System Events" to display dialog "Password to purge virtual memory:" default answer "" with hidden answer\' -e \'text returned of result\' 2>/dev/null)" && echo "$pw"\n')

        # call the purge command while prompting for the password (MacOS for some reason requires high privileges on the Purge command)
        # (with 4 retries in case of mac 'error initializing audit plugin sudoers_audit' error)
        os.environ['SUDO_ASKPASS'] = './mac_prompt_pw.sh'
        os.system('chmod +x ./mac_prompt_pw.sh')
        retry_purge = 0
        while retry_purge < 5:
            if os.system('sudo -A /usr/sbin/purge') == 0:
                break
            else:
                print('Retry purge')
                retry_purge += 1
        if retry_purge == 5:
            ConsoleColors.print_error('Could not purge virtual memory after 5 tries, skipping')
        os.remove('./mac_prompt_pw.sh')

        #
        print('Cleared virtual memory')
